//
//  CRF++ -- Yet Another CRF toolkit
//
//  $Id: encoder.h 1588 2007-02-12 09:03:39Z taku $;
//
//  Copyright(C) 2005-2007 Taku Kudo <taku@chasen.org>
//
#ifndef CRFPP_ENCODER_H_
#define CRFPP_ENCODER_H_

#include "common.h"
#include <vector>
#include <string>
#include "tagger.h"
#include "feature_index.h"

using namespace std;

namespace CRFPP {

class Encoder {
private:
	whatlog what_;
public:
	size_t         freq; //use features that occuer no less than INT(default 1)
	size_t         maxitr; //set INT for max iterations in LBFGS routine(default 10k)
	double         C; //set FLOAT for cost parameter(default 1.0)
	double         eta; //set FLOAT for termination criterion(default 0.0001)
	bool           textmodel;
	unsigned short thread_num;
	unsigned short shrinking_size;//set INT for number of iterations variable needs to be optimal before considered for shrinking. (default 20)
	std::string salgo; //crf-l1 or crf-l2, default crf-l2
	int            algorithm;

	double qGaussianVariance;

	std::vector<TaggerImpl* > labelx;
	std::vector<TaggerImpl* > devx;

	vector<double>   mu;
	vector<double>   grad_mu; //gradient for each constraint, sum up to iteration t

	vector <double>  alpha;
	vector <int>     alpha_count;
	vector<double>   grad_alpha; //gradient for each feature, sum up to iteration t

	//vector<PRConstraint> constraintsList;
	EncoderFeatureIndex* feature_index;

	bool ada_grad;

public:
   Encoder()
	{
		freq           = 1;
		maxitr        = 100000; //default: 100000
		C              = 1.0;
		eta            = 0.0001;
		textmodel      = false;
		thread_num         = 1;
		shrinking_size = 20;
		salgo = "crf";

		feature_index = NULL;

		ada_grad = true;
	}

   ~Encoder() {
	    for (std::vector<TaggerImpl *>::iterator it = labelx.begin(); it != labelx.end(); ++it)
	   		delete *it;
		for (std::vector<TaggerImpl *>::iterator it = devx.begin(); it != devx.end(); ++it)
			delete *it;
   }

  void RebuildAlpha();

  //batch learning
  bool learn(const char * modelfile);

  bool SGD_learn(string modelfile, double learningRate, double decayRate,
  		int numIterations);

  bool singleLabelLikelihood(TaggerImpl* x, EncoderFeatureIndex *feature_index,
  	  double *alpha, vector<int> &alpha_count, vector<double> &expected,
  	  double rate, double &value, double C, int n, int iter);

  double calculateAcc(const vector<TaggerImpl*> data);

  const char* what() { return what_.str(); }
};

}
#endif
