//
//  CRF++ -- Yet Another CRF toolkit
//
//  $Id: feature_index.cpp 1587 2007-02-12 09:00:36Z taku $;
//
//  Copyright(C) 2005-2007 Taku Kudo <taku@chasen.org>
//
#include <iostream>
#include <fstream>
#include <cstring>
#include <set>

#include "common.h"
#include "feature_index.h"
#include "../Parsing/Utils.h"

namespace CRFPP {

static inline char *read_ptr(char **ptr, size_t size) {
  char *r = *ptr;
  *ptr += size;
  return r;
}

template <class T> static inline void read_static(char **ptr,
                                                  T *value) {
  char *r = read_ptr(ptr, sizeof(T));
  memcpy(value, r, sizeof(T));
}

int DecoderFeatureIndex::getID(const char *key) {
  return da_.exactMatchSearch<Darts::DoubleArray::result_type>(key);
}

int DecoderFeatureIndex::getfID(const char *key) {
  return da_.exactMatchSearch<Darts::DoubleArray::result_type>(key);
}

int EncoderFeatureIndex::addID(const char *key, int fsize)
{
  std::map <std::string, std::pair<int, unsigned int> >::iterator it = dic_.find(key);
  if (it == dic_.end()) {
    dic_.insert(std::make_pair<std::string, std::pair<int, unsigned int> >
                (key, std::make_pair<int, unsigned int>(maxid_, 1)));
    int n = maxid_;
	maxid_ += fsize;
	return n;
  } else {
    it->second.second++;
    return it->second.first;
  }
  return -1;
}

int EncoderFeatureIndex::getID(const char *key)
{
  std::map <std::string, std::pair<int, unsigned int> >::iterator it = dic_.find(key);
  if (it == dic_.end()) {
    dic_.insert(std::make_pair<std::string, std::pair<int, unsigned int> >
                (key, std::make_pair<int, unsigned int>(maxid_, 1)));
    int n = maxid_;
    int fsize = (key[0] == 'B' ? y_.size() * y_.size(): y_.size());

	maxid_ += fsize;
	return n;
  } else {
    it->second.second++;
    return it->second.first;
  }
  return -1;
}

int EncoderFeatureIndex::getfID(const char *key)
{
  std::map <std::string, std::pair<int, unsigned int> >::iterator it = dic_.find(key);
  if (it == dic_.end())
     return -1;
  else
    return it->second.first;
}

bool EncoderFeatureIndex::open(const char *filename1) {
  return openTagSet(filename1);
}

bool EncoderFeatureIndex::openTagSet(const char *file) {
  std::ifstream ifs(file);
  CHECK_FALSE(ifs) << "no such file or directory: " << file ;

  char  line[8192];
  char* column[1024];
  size_t max_size = 0;
  std::set<std::string> candset;

  int lno = 1;
  while (ifs.getline(line, sizeof(line))) {
    if (line[0] == '\0' || line[0] == ' ' || line[0] == '\t') continue;
    size_t size = tokenize2(line, "\t ", column, 1024);
    //if (max_size == 0) 
	max_size = size;
    /*CHECK_FALSE(max_size == size)
        << "inconsistent column size: "
        << max_size << " " << size << " " << file<< ":"<<lno<<" "<<line;*/
    lno++;
	string y = column[max_size-1];
	if(candset.find(y) == candset.end())
		candset.insert(y);
  }

  y_.clear();
  for (std::set<std::string>::iterator it = candset.begin(); it != candset.end(); ++it)
  {
	  lb2i[*it] = y_.size();
	  i2lb[y_.size()] = *it;
	  y_.push_back(this->strdup(it->c_str()));
  }

  ifs.close();

  return true;
}

bool EncoderFeatureIndex::opendata(vector<vector<string> > crfdata) {

	char* column[1024];

  size_t max_size = 0;
  std::set<std::string> candset;

  int lno = 1;
  for(int i=0; i<crfdata.size(); i++)
  {
	  for(int j=0; j<crfdata[i].size(); j++)
	  {
		size_t len = std::strlen(crfdata[i][j].c_str());
		char *q = new char[len+1];
		std::strcpy(q, crfdata[i][j].c_str());
		size_t size = tokenize2(q, "\t ", column, 1024);
		
		max_size = size;
		lno++;
		string y = column[max_size-1];
		if(candset.find(y) == candset.end())
			candset.insert(y);
	  }
  }

  y_.clear();
  for (std::set<std::string>::iterator it = candset.begin();
       it != candset.end(); ++it)
  {
	  lb2i[*it] = y_.size();
	  i2lb[y_.size()] = *it;
	  y_.push_back(this->strdup(it->c_str()));
  }

  return true;
}

bool DecoderFeatureIndex::open(const char *filename1) {
  CHECK_FALSE(mmap_.open(filename1)) << mmap_.what();

  char *ptr = mmap_.begin();

  read_static<double>(&ptr, &cost_factor_);
  read_static<int>(&ptr, &windowsize);
  read_static<bool>(&ptr, &zero_order);
  read_static<int>(&ptr, &max_span_length);

  read_static<int>(&ptr, &word_vec_size);
  read_static<int>(&ptr, &embedding_label_size);

  read_static<unsigned int>(&ptr, &maxid_);

  unsigned int y_str_size;
  read_static<unsigned int>(&ptr, &y_str_size);
  char *y_str = read_ptr(&ptr, y_str_size);
  size_t pos = 0;
  while (pos < y_str_size) {
    y_.push_back(y_str + pos);
    while (y_str[pos++] != '\0') {}
  }

  for(int i=0; i<y_.size(); i++) {
	  string y = y_[i];
	  lb2i[y] = i;
	  i2lb[i] = y;
  }

    y_str_size = 0;
    read_static<unsigned int>(&ptr, &y_str_size);
    y_str = read_ptr(&ptr, y_str_size);
    pos = 0;
    while (pos < y_str_size) {
      polar_ystr.push_back(y_str + pos);
      while (y_str[pos++] != '\0') {}
    }
    polar_ysize_ = polar_ystr.size();

    y_str_size = 0;
    read_static<unsigned int>(&ptr, &y_str_size);
    y_str = read_ptr(&ptr, y_str_size);
	pos = 0;
	while (pos < y_str_size) {
	  intensity_ystr.push_back(y_str + pos);
	  while (y_str[pos++] != '\0') {}
	}
	intensity_ysize_ = intensity_ystr.size();

  //transition states
  	unsigned int tran_size = 0;
  	read_static<unsigned int>(&ptr, &tran_size);
  	for (int i = 0; i < tran_size; ++i) {
  		int yj = 0;
  		int yi = 0;
  		read_static<int>(&ptr, &yj);
  		read_static<int>(&ptr, &yi);
  		transition_states[make_pair(yj, yi)] = 1;
  	}

  unsigned int dsize = 0;
  read_static<unsigned int>(&ptr, &dsize);
  da_.set_array(ptr);
  ptr += dsize;

  alpha_ = reinterpret_cast<double *>(ptr);
  ptr += sizeof(alpha_[0]) * maxid_;

  CHECK_FALSE(ptr == mmap_.end()) << "model file is broken: " << filename1;

  return true;
}

void EncoderFeatureIndex::shrink(size_t freq) {
  if (freq <= 1) return;

  std::map<int, int> old2new;
  int new_maxid = 0;

  for (std::map<std::string, std::pair<int, unsigned int> >::iterator
           it = dic_.begin(); it != dic_.end();) {
    const std::string &key = it->first;

    if (it->second.second >= freq) {
      old2new.insert(std::make_pair<int, int>(it->second.first, new_maxid));
      it->second.first = new_maxid;
      if (key[0] == 'B') {
    	  	  new_maxid += y_.size() * y_.size();
      } else if (key[0] == 'S') {
    	      new_maxid += y_.size();
      } else if (key[0] == 'P') {
    	      new_maxid += polar_ystr.size();
      } else if (key[0] == 'I') {
	      new_maxid += intensity_ystr.size();
      }
      ++it;
    } else {
      dic_.erase(it++);
    }
  }

  feature_cache_.shrink(&old2new);

  maxid_ = new_maxid;

  return;
}

void FeatureIndex::clear() {
	ClearFeatureMap();

  char_freelist_.free();
  feature_cache_.clear();
  feature_value_cache_.clear();

  for (size_t i = 0; i < thread_num_; ++i) {
    node_freelist_[i].free();
    path_freelist_[i].free();

    segnode_freelist_[i].free();
  }
}

bool EncoderFeatureIndex::save(const char *filename, bool textmodelfile) {
	std::ofstream bofs;
	bofs.open(filename, OUTPUT_MODE);
    CHECK_FALSE(bofs) << "open failed: " << filename;

	bofs.write(reinterpret_cast<char *>(&cost_factor_), sizeof(cost_factor_));
	bofs.write(reinterpret_cast<char *>(&windowsize), sizeof(windowsize));
	bofs.write(reinterpret_cast<char *>(&zero_order), sizeof(zero_order));
	bofs.write(reinterpret_cast<char *>(&max_span_length), sizeof(max_span_length));

	bofs.write(reinterpret_cast<char *>(&word_vec_size), sizeof(word_vec_size));
	bofs.write(reinterpret_cast<char *>(&embedding_label_size), sizeof(embedding_label_size));

	bofs.write(reinterpret_cast<char *>(&maxid_), sizeof(maxid_));

	//label info
	std::string y_str = "";
	for (size_t i = 0; i < y_.size(); ++i) {
	   y_str += std::string(y_[i]);
	   y_str += '\0';
	}
	unsigned int size = y_str.size();
	bofs.write(reinterpret_cast<char *>(&size),  sizeof(size));
	bofs.write(const_cast<char *>(y_str.data()), y_str.size());

	//attribute info
	y_str = "";
	for (size_t i = 0; i < polar_ystr.size(); ++i) {
		y_str += polar_ystr[i];
		y_str += '\0';
	}
	size = y_str.size();
	bofs.write(reinterpret_cast<char *>(&size),  sizeof(size));
	bofs.write(const_cast<char *>(y_str.data()), y_str.size());

	y_str = "";
	for (size_t i = 0; i < intensity_ystr.size(); ++i) {
		y_str += intensity_ystr[i];
		y_str += '\0';
	}
	size = y_str.size();
	bofs.write(reinterpret_cast<char *>(&size),  sizeof(size));
	bofs.write(const_cast<char *>(y_str.data()), y_str.size());

	//transition states
	unsigned int tran_size = transition_states.size();
	bofs.write(reinterpret_cast<char *>(&tran_size),  sizeof(tran_size));
	for (map<pair<int, int>, int>::iterator it = transition_states.begin();
			it != transition_states.end(); ++it) {
		int yj = it->first.first;
		int yi = it->first.second;
		bofs.write(reinterpret_cast<char *>(&yj), sizeof(yj));
		bofs.write(reinterpret_cast<char *>(&yi), sizeof(yi));
	}

	//feature dic_
	std::vector <char *> key;
	std::vector <int>    val;
	for (std::map<std::string, std::pair<int, unsigned int> >::iterator it = dic_.begin(); it != dic_.end(); ++it) {
		key.push_back(const_cast<char *>(it->first.c_str()));
		val.push_back(it->second.first);
	}
	Darts::DoubleArray da;
	CHECK_FALSE(da.build(key.size(), &key[0], 0, &val[0]) == 0) << "cannot build double-array";
	unsigned int dsize = da.unit_size() * da.size();
	bofs.write(reinterpret_cast<char *>(&dsize), sizeof(dsize));
	bofs.write(reinterpret_cast<const char *>(da.array()), dsize);

    //feature param
    for (size_t i  = 0; i < maxid_; ++i) {
        double alpha = alpha_[i];
        bofs.write(reinterpret_cast<char *>(&alpha), sizeof(alpha));
    }

    bofs.close();

    if (textmodelfile) {
       std::string filename2 = filename;
       filename2 += ".txt";

       std::ofstream tofs(filename2.c_str());

       CHECK_FALSE(tofs) << " no such file or directory: " << filename2;

       // header
       tofs << "cost-factor: " << cost_factor_ << std::endl;
       tofs << "maxid: "       << maxid_ << std::endl;
       tofs << "windowsize: "  << windowsize <<endl;
       tofs << std::endl;

	   // y
	  for (size_t i = 0; i < y_.size(); ++i)
	      tofs << y_[i] << std::endl;
	  tofs << std::endl;

      // dic
      for (std::map<std::string, std::pair<int, unsigned int> >::iterator
             it = dic_.begin();
          it != dic_.end(); ++it) {
          tofs << it->first << std::endl;
      }

      tofs << std::endl;
      tofs.setf(std::ios::fixed, std::ios::floatfield);
      tofs.precision(16);

      int ysize_ = ysize();
      // !!!!! for structured perceptron
      if (ysize_ == 0) ysize_ = 1;
      for (std::map<std::string, std::pair<int, unsigned int> >::iterator
                it = dic_.begin(); it != dic_.end(); ++it) {
    	      string fstr = it->first;
    	      int fid = it->second.first;
    	      if (fstr[0] == 'B') {
    	    	      for (int i = 0; i < ysize_*ysize_;++i) {
    	    	    	  	  tofs << fstr <<" "<< fid+i<<" "<< i << " " << get_alpha(fid+i) << std::endl;
    	    	      }
    	      } else {
    	    	      for (int i = 0; i < ysize_; ++i) {
    	    	    	 	  tofs << fstr <<" "<< fid+i<<" "<< i << " " << get_alpha(fid+i) << std::endl;
    	    	      }
    	      }
       }
	   tofs << std::endl;
   }

   return true;
}

char *FeatureIndex::strdup(const char *p) {
  size_t len = std::strlen(p);
  char *q = char_freelist_.alloc(len+1);
  std::strcpy(q, p);
  return q;
}

void FeatureIndex::calcCost(Node *n) {
  n->cost = 0.0;

#define ADD_COST(T, A)                                                  \
  { T c = 0;                                                            \
    for (int *f = n->fvector; *f != -1; ++f) c += (A)[*f + n->y];       \
    n->cost =cost_factor_ *(T)c; }

  if (alpha_float_) ADD_COST(float,  alpha_float_)
    else             ADD_COST(double, alpha_);
#undef ADD_COST
}

void FeatureIndex::calcCost(Path *p) {
  p->cost = 0.0;

#define ADD_COST(T, A)                                          \
  { T c = 0.0;                                                  \
    for (int *f = p->fvector; *f != -1; ++f)                    \
      c += (A)[*f + p->lnode->y * y_.size() + p->rnode->y];     \
    p->cost =cost_factor_*(T)c; }

  if (alpha_float_) ADD_COST(float,  alpha_float_)
    else             ADD_COST(double, alpha_);
#undef ADD_COST
}

void FeatureIndex::LoadFeatureMap(string filename) {
	feature_map.clear();
	ifstream infile(filename.c_str(), ios::in);
	string str;
	while (getline(infile, str)) {
		int index = str.find('\t');
		string key = str.substr(0, index);
		string value = str.substr(index+1);
		vector<string> fields;
		Utils::Split(value, ' ', fields);
		feature_map[key] = fields;
	}
	infile.close();
}

void FeatureIndex::LoadEmbeddings(string infilename) {
	word_lookup_table.clear();
	word_embeddings.clear();

	ifstream infile(infilename.c_str(), ios::in);
	string str;
	while (getline(infile, str)) {
		int index = str.find('\t');
		string word = str.substr(0, index);
		string values = str.substr(index+1);
		vector<string> fields;
		Utils::Split(values, ' ', fields);
		vector<float> vec(fields.size());
		for (int i = 0; i <fields.size(); ++i) vec[i] = atof(fields[i].c_str());
		word_lookup_table[word] = word_lookup_table.size();
		word_embeddings.push_back(vec);
	}
	infile.close();
}

void FeatureIndex::InitEmbeddingFeatures() {
	if (word_embeddings.size() == 0) return;

	word_vec_size = word_embeddings[0].size();
	for (int i = 0; i < word_vec_size; ++i) {
		string feature = "Embedding_"+Utils::int2string(i);
		addID(feature.c_str(), embedding_label_size);
	}
}

void FeatureIndex::ClearWordEmbeddings() {
	word_embeddings.clear();
	word_lookup_table.clear();
}

void FeatureIndex::ClearFeatureMap() {
	feature_map.clear();
}
}
