/*
 * semi_encoder.h
 *
 *  Created on: Feb 15, 2014
 *      Author: bishan
 */

#ifndef SEMI_ENCODER_H_
#define SEMI_ENCODER_H_
#include "semi_tagger.h"

namespace CRFPP {

class SemiCRFEncoder {
public:
	size_t         maxitr; //set INT for max iterations in LBFGS routine(default 10k)
	double         C; //set FLOAT for cost parameter(default 1.0)
	double         eta; //set FLOAT for termination criterion(default 0.0001)
	bool           textmodel;
	unsigned short thread_num;
	int            freq;
	bool           loss_aware;

	std::vector<Tagger* > labelx;
	std::vector<Tagger* > devx;

	vector <double>  alpha;
	vector <int>     alpha_count;
	vector<double>   grad_alpha; //gradient for each feature, sum up to iteration t

	//vector<PRConstraint> constraintsList;
	EncoderFeatureIndex* feature_index;

	bool ada_grad;

	public:
	SemiCRFEncoder() {
		maxitr        = 100000; //100000
		C              = 1;
		eta            = 0.0001;
		textmodel      = false;
		thread_num     = 1;
		ada_grad = true;
		freq           = 1;
		loss_aware = false;
	}

	~SemiCRFEncoder() {
		for (std::vector<Tagger *>::iterator it = labelx.begin(); it != labelx.end(); ++it)
			delete *it;
		for (std::vector<Tagger *>::iterator it = devx.begin(); it != devx.end(); ++it)
			delete *it;
	}

	double Eval(const vector<Tagger*> data);

	//initialize alpha, mu ... for feature index
	void RebuildAlpha();

	bool LBFGS_optimizer(const std::vector<Tagger* > &x,
	      EncoderFeatureIndex *feature_index, double *alpha,
	      vector<double>& expected, double &value,
	      size_t maxitr, double C, double eta,
	      bool orthant);

	bool SemiLabelLikelihood(const std::vector<Tagger* > &x,
	      EncoderFeatureIndex *feature_index, double *alpha,
	      vector<double>& expected, double &value,
	      size_t maxitr, double C, double eta,
	      unsigned short thread_num, bool orthant);

	bool learn(const char * modelfile);

	bool SGD_learn(string modelfile, double learningRate, double decayRate,
			int numIterations);

	bool singleLabelLikelihood(Tagger* x, EncoderFeatureIndex *feature_index,
		  double *alpha, vector<int> &alpha_count, vector<double> &expected,
		  double rate, double &value, double C, int n, int iter);

	double calculateAcc(const vector<Tagger*> data);
};

} /* namespace CRFPP */
#endif /* SEMI_ENCODER_H_ */
