/*
 * semi_tagger.cpp
 *
 *  Created on: Feb 15, 2014
 *      Author: bishan
 */

#include "semi_tagger.h"

//#define DEBUG_

namespace CRFPP {

int SemiTagger::get_lb2i(string str)
{
	return feature_index_->get_lb2i(str);
}
string SemiTagger::get_i2lb(int i)
{
	string str = feature_index_->ystr(i);
	return str;
}

void SemiTagger::initTagger() {
	//initialize node_(segmentsEnd), node_beta_(segmentsStart)
	node_.resize(segmentsEnd.size());
	node_beta_.resize(segmentsStart.size());
	for(int i =0; i<segmentsEnd.size(); i++) {
	  for(int j=0; j<segmentsEnd[i].size(); j++) {
		  vector<SegmentNode*> nodes;
		  nodes.resize(ysize());
		  node_[i].push_back(nodes);
	  }
	}

	for(int i=0; i<segmentsStart.size(); i++) {
	  for(int j=0; j<segmentsStart[i].size(); j++) {
		  vector<SegmentNode*> nodes_beta;
		  nodes_beta.resize(ysize());
		  node_beta_[i].push_back(nodes_beta);
	  }
	}

	//initialize alpha, beta
	alpha.resize(size()+1);
	for(int i=0; i<(int)alpha.size(); i++) {
		alpha[i].resize(ysize_);
		fill(alpha[i].begin(), alpha[i].end(), 0);
	}
	beta.resize(size()+1);
	for(int i=0; i<(int)beta.size(); i++) {
		beta[i].resize(ysize_);
		fill(beta[i].begin(), beta[i].end(), 0);
	}

	alpha_w_loss.resize(size()+1);
	for(int i=0; i<(int)alpha_w_loss.size(); i++) {
		alpha_w_loss[i].resize(ysize_);
		fill(alpha_w_loss[i].begin(), alpha_w_loss[i].end(), 0);
	}

	beta_w_loss.resize(size()+1);
	for(int i=0; i<(int)beta_w_loss.size(); i++) {
		beta_w_loss[i].resize(ysize_);
		fill(beta_w_loss[i].begin(), beta_w_loss[i].end(), 0);
	}

	yycost.resize(ysize_);
	for(int i=0; i<(int)yycost.size(); i++) {
		yycost[i].resize(ysize_);
		fill(yycost[i].begin(), yycost[i].end(), 0);
	}
}

void SemiTagger::BuildUnitBasedTrueSegmentation() {
	int i = 0;
	while (i < span_units.size()) {
		if (span_units[i].label == "O") {
			true_segmentation.push_back(span_units[i]);
			i++;
		} else {
			vector<Span> entity_unit_spans;
			entity_unit_spans.push_back(span_units[i]);
			int s = span_units[i].start;
			i++;
			while (i < span_units.size() && IsInSameSegment(s, span_units[i].end)) {
				entity_unit_spans.push_back(span_units[i]);
				i++;
			}

			Span wholespan(entity_unit_spans[0].start,
					entity_unit_spans[entity_unit_spans.size()-1].end,entity_unit_spans[0].label);
			true_segmentation.push_back(wholespan);
		}
	}

	//map token to index of true segmentation
	token_to_truesegmentation.resize(size(), 0);
	for (int i = 0; i < true_segmentation.size(); ++i) {
		for (int j = true_segmentation[i].start; j <= true_segmentation[i].end; ++j) {
			token_to_truesegmentation[j] = i;
		}
	}
}

void SemiTagger::BuildTrueSegmentation(int max_len) {
	int prev_end = -1;
	int i = 0;
	while (i < sent->entities.size()) {
		// gap between entities
		for (int j = prev_end+1; j < sent->entities[i].start; ++j) {
			true_segmentation.push_back(Span(j, j, "O"));
		}
		// entities
		if (sent->entities[i].len() > max_len) {
			for (int j = sent->entities[i].start; j <= sent->entities[i].end; ++j) {
				true_segmentation.push_back(Span(j, j, sent->entities[i].label));
			}
		} else {
			true_segmentation.push_back(Span(sent->entities[i].start, sent->entities[i].end, sent->entities[i].label));
		}
		prev_end = sent->entities[i].end;
		i++;
	}
	for (int j = prev_end+1; j < size(); ++j) {
		true_segmentation.push_back(Span(j, j, "O"));
	}

	//map token to index of true segmentation
	token_to_truesegmentation.resize(size(), 0);
	for (int i = 0; i < true_segmentation.size(); ++i) {
		for (int j = true_segmentation[i].start; j <= true_segmentation[i].end; ++j) {
			token_to_truesegmentation[j] = i;
		}
	}
}

bool SemiTagger::IsTrueSegment(int start, int end) {
	// check token to segment map
	int seg_index = token_to_truesegmentation[start];
	// check boundaries
	if (start > 0 && token_to_truesegmentation[start-1] == seg_index) return false;
	if (end < size()-1 && token_to_truesegmentation[end+1] == seg_index) return false;

	// check inside
	return (token_to_truesegmentation[end] == seg_index);
}

bool SemiTagger::IsInSameSegment(int i, int j) {
	return token_to_entities[i] == token_to_entities[j];
}

bool SemiTagger::SentenceToSemiTagger(Sentence *psent,
		CRFPP::FeatureIndex *feature_index, bool train, bool useparse) {
	open(feature_index);

	sent = psent;
	size_t s = sent->size();
	if (s == 0) {
		return false;
	}

	x_.resize(s);
	answer_.resize(s);
	result_.resize(s, 0);

	for(int t=0; t<sent->token_labels.size(); t++) {
		x_[t].push_back(sent->words[t]);
		string label = sent->token_labels[t];
		answer_[t] = feature_index_->yID(label);
	}

	// candidate text spans
	for (int i = 0; i < s; ++i) {
		vector<int> starts;
		segmentsStart.push_back(starts);
		vector<int> ends;
		segmentsEnd.push_back(ends);
	}

	if (train) {
		if (useparse && sent->penntree != NULL) {
			token_to_entities.resize(s, -1);
			for (int i = 0; i < sent->entities.size(); ++i) {
				for (int k = sent->entities[i].start; k <= sent->entities[i].end; ++k) {
					token_to_entities[k] = i;
				}
			}

			// adjust the units to have consistent labels
			span_units.clear();
			for (int i = 0; i < sent->penntree->units.size(); ++i) {
				int start = sent->penntree->units[i].start;
				int end = sent->penntree->units[i].end;
				string label = feature_index_->ystr(answer_[start]);
				bool valid = true;
				for (int j = start; j <= end; ++j) {
					if (feature_index_->ystr(answer_[j]) != label) {
						valid = false;
					}
				}
				if (valid) {
					span_units.push_back(Span(start, end, label));
				} else {
					for (int j = start; j <= end; ++j) {
						span_units.push_back(Span(j, j, feature_index_->ystr(answer_[j])));
					}
				}
			}

			BuildUnitBasedTrueSegmentation();

			// generate candidate spans
			genSegmentUnitCandidates();
		} else {
			int max_length = -1;
			for (int i = 0; i < sent->entities.size(); ++i) {
				if (sent->entities[i].len() > max_length) {
					max_length = sent->entities[i].len();
				}
			}
			if (max_length < 0) max_length = 1;
			BuildTrueSegmentation(max_length);
			genSegmentCandidates(max_length);
		}
	} else { //test
//		if (useparse && sent->penntree != NULL) {
//			span_units.clear();
//			for (int i = 0; i < sent->penntree->units.size(); ++i) {
//				int start = sent->penntree->units[i].start;
//				int end = sent->penntree->units[i].end;
//				span_units.push_back(Span(start, end, "O"));
//			}
//			genTestSegmentUnitCandidates(feature_index_->max_span_length);
//		} else {
			genSegmentCandidates(feature_index_->max_span_length);
		//}
	}

	// allocate nodes
	initTagger();

	// build features
	feature_index_->buildSemiCRFFeatures(this, train);

	return true;
}

double SemiTagger::gradient(double *expected) {
   if (x_.empty()) return 0.0;

   // reallocate node structure
   feature_index_->buildSemiCRFTagger(this);

#ifdef DEBUG_
   cout<<"lambda: "<<endl;
  	for (int i = 0; i < 20; ++i) {
  		 cout<<feature_index_->get_alpha(i)<<" ";
  	}
  	cout<<endl;
#endif

   ComputeCostWithLoss(feature_index_->loss_aware); //calculate node cost and edge cost

   //forwardbackward(); //calculate alpha, beta

	calcAlphawithLoss();
	calcBetawithLoss();

	Z_ = 0.0;
	for (size_t j = 0; j < ysize_; ++j)
		Z_ = logsumexp(Z_, alpha_w_loss[x_.size()][j], j == 0);

    //calcExpectation(expected); //calculate expected
	calcExpectationWithLoss(expected);

#ifdef DEBUG_
    cout<<"answer_"<<endl;
	for (int i = 0; i < x_.size(); ++i) {
		cout<<sent->token_labels[i]<<":"<<answer_[i]<<" ";
	}
	cout<<endl;

	cout<<sent->toString()<<endl;
    cout<<"expected before: "<<endl;
   	for (int i = 0; i <10; ++i) {
   		cout<<expected[i]<<" ";
   	}
   	cout<<endl;
#endif

    double s = 0.0;
   	for (int i = 0; i < true_segmentation.size(); ++i) {
   		int start = true_segmentation[i].start;
   		int end = true_segmentation[i].end;
   		SegmentNode* node = GetSegmentNode(start, end, answer_[start]);
		for (int *f = node->fvector; *f != -1; ++f) {
			expected[*f + node->y] -= 1.0;
		}
		double *fv = node->embedding_word_vec;
		for (int k = 0; k < feature_index_->word_vec_size; ++k) {
			double value = *fv;
			expected[k*feature_index_->embedding_label_size + node->y] -= value;
			++fv;
		}

		s += node->cost;  // UNIGRAM cost
		assert(node->y == answer_[start]);
		if(i > 0) {
			int prey = answer_[start-1];
		    expected[feature_index_->edgeid_+ prey * ysize_ + node->y] -= 1.0;
		    s += getTransitionCost(prey, node->y);  // BIGRAM COST
		}
   	}
/*   	vector<int> segment_y;
   for (int i = 0; i < x_.size(); ++i) {//calculate alpha ending at position i with label j
   		for(int ns=0; ns<numSegmentsEndAt(i); ns++) {
   			int start = segmentsEnd[i][ns];
   			int end = i;
   			if (IsTrueSegment(start, end)) {
   				SegmentNode* node = GetSegmentNode(start, end, answer_[i]);
   				for (int *f = node->fvector; *f != -1; ++f) {
   					expected[*f + node->y] -= 1.0;
   				}
   				s += node->cost;  // UNIGRAM cost
   				assert(node->y == answer_[i]);
   				if(segment_y.size() > 0) {
   					int prey = segment_y[segment_y.size()-1];
   					// only model different label transition
   					//if (node->y != answer_[i-1]) {
   					//cout<<"edge "<<i<<" "<<prey<<","<<node->y<<":"<<getTransitionCost(prey, node->y)<<endl;
   						expected[feature_index_->edgeid_+ prey * ysize_ + node->y] -= 1.0;
   						s += getTransitionCost(prey, node->y);  // BIGRAM COST

   						//cout<<feature_index_->edgeid_+ prey * ysize_ + node->y
   						//	<<" "<<expected[feature_index_->edgeid_+ prey * ysize_ + node->y]<<endl;

   					//}
				}

   				segment_y.push_back(node->y);
   			}
   		}
   }
   //cout<<endl;
*/
   //viterbi();  // call for eval()
#ifdef DEBUG_
    cout<<"expected for gradient: "<<endl;
   	for (int i = 0; i <10; ++i) {
   		cout<<expected[i]<<" ";
   	}
   	cout<<endl;
   cout<<"Z_ "<<Z_<<" "<<s<<endl;
#endif
   return Z_ - s;
 }

 double SemiTagger::toprob(SegmentNode *n) { //p(x[s:t]|x)
	 if(n->start > 0) {
		double c = 0.0;
		for(int prey=0; prey<(int)ysize_; prey++) {
			c = logsumexp(c, alpha[n->start-1][prey] + getTransitionCost(prey,n->y), prey == 0);
		}
		c += n->cost + beta[n->end][n->y];
		return std::exp(c - Z_);
	 } else {
		return std::exp(n->cost + beta[n->end][n->y] - Z_);
	 }
  }

 double SemiTagger::toyyprob(SegmentNode* n, int prey) //[,i-1:prey][i,:y]
 {
	 double c = alpha[n->start-1][prey] + getTransitionCost(prey,n->y) + n->cost + beta[n->end][n->y];
	 return std::exp(c - Z_);
 }

 void SemiTagger::calcExpectationWithLoss(double *expected)
  {
 	for (int i = 0; i < x_.size(); ++i) {//calculate alpha ending at position i with label j
 		for(int ns=0; ns<numSegmentsEndAt(i); ns++) {
 			for (size_t k = 0; k < node_[i][ns].size(); ++k) {
 				SegmentNode* node = node_[i][ns][k];
 				int yp = node->y;
 				double c = 0.0;
 				if (node->start-1 < 0) {
 					c = node->cost_w_loss + beta_w_loss[i][yp];
 				} else {
 					for(int prey=0; prey<(int)ysize_; prey++) {
 						c = logsumexp(c, alpha_w_loss[node->start][prey] + getTransitionCost(prey, yp), prey==0);
 					}
 					c += node->cost_w_loss + beta_w_loss[i][yp];
 				}

 				node->prob = std::exp(c - Z_);

 				for (int *f = node->fvector; *f != -1; ++f) {
 					expected[*f + node->y] += node->prob;
 				}
 				double *fv = node->embedding_word_vec;
				for (int v = 0; v < feature_index_->word_vec_size; ++v) {
					double value = *fv;
					expected[v * feature_index_->embedding_label_size + node->y] += value * node->prob;
					++fv;
				}

 				// edge cost
 				if (node->start-1 >= 0) {
 					for(int prey=0; prey<(int)ysize_; prey++) {
 						//if (prey == yp) continue;
 						double prob = std::exp(alpha_w_loss[node->start][prey] + node->cost_w_loss + getTransitionCost(prey, yp) + beta_w_loss[i][yp] - Z_);
 						expected[feature_index_->edgeid_+prey*ysize_+yp] += prob;
 					}
 				}
 			}
 		}
 	}
  }

 void SemiTagger::calcExpectation(double *expected)
 {
	for (int i = 0; i < x_.size(); ++i) {//calculate alpha ending at position i with label j
		for(int ns=0; ns<numSegmentsEndAt(i); ns++) {
			for (size_t k = 0; k < node_[i][ns].size(); ++k) {
				SegmentNode* node = node_[i][ns][k];
				int yp = node->y;
				double c = 0.0;
				if (node->start-1 < 0) {
					c = node->cost + beta[i][yp];
				} else {
					for(int prey=0; prey<(int)ysize_; prey++) {
						c = logsumexp(c, alpha[node->start][prey] + getTransitionCost(prey, yp), prey==0);
					}
					c += node->cost + beta[i][yp];
				}
				node->prob = std::exp(c - Z_);
				for (int *f = node->fvector; *f != -1; ++f) {
					expected[*f + node->y] += node->prob;
				}

				// edge cost
				if (node->start-1 >= 0) {
					for(int prey=0; prey<(int)ysize_; prey++) {
						//if (prey == yp) continue;
						double prob = std::exp(alpha[node->start][prey] + node->cost + getTransitionCost(prey, yp) + beta[i][yp]-Z_);
						expected[feature_index_->edgeid_+prey*ysize_+yp] += prob;
					}
				}
			}
		}
	}

/*	map<int, double> tagger_expected;
	int edgeid = feature_index_->edgeid_;
	for (int prey = 0; prey < ysize_; ++prey) {
		for (int py = 0; py < ysize_; ++py) {
			tagger_expected[edgeid + prey*ysize_+py] = LOG0;
		}
	}
	for (int i = 0; i < x_.size(); ++i) {//calculate alpha ending at position i with label j
		for(int ns=0; ns<numSegmentsEndAt(i); ns++) {
			for (size_t yp = 0; yp < ysize_; ++yp) {
				SegmentNode* node = node_[i][ns][yp];
				for (int *f = node->fvector; *f != -1; ++f) {
					tagger_expected[*f + node->y] = LOG0;
				}
			}
		}
	}
	for (int i = 0; i < x_.size(); ++i) {//calculate alpha ending at position i with label j
		for(int ns=0; ns<numSegmentsEndAt(i); ns++) {
			for (size_t yp = 0; yp < ysize_; ++yp) {
				SegmentNode* node = node_[i][ns][yp];
				for (int *f = node->fvector; *f != -1; ++f) {
					int fid = *f + node->y;
					if (node->start-1 < 0) {
						tagger_expected[fid] = logsumexp(tagger_expected[fid], node->cost + beta[i][yp]);
						//cout.precision(15);
						//cout<<node->start<<" "<<node->end<<" "<<fid<<" "<< yp <<" "<< tagger_expected[fid]<<endl;
					} else {
						for(int prey=0; prey<(int)ysize_; prey++) {
							tagger_expected[fid] = logsumexp(tagger_expected[fid],
									alpha[node->start][prey] + node->cost + yycost[prey][yp] + beta[i][yp]);
							//cout.precision(15);
							//cout<<node->start<<" "<<node->end<<" "<<fid<<" "<< yp <<" "<< tagger_expected[fid]<<endl;
						}
					}
				}

				// edge cost
				if (node->start-1 >= 0) {
					for(int prey=0; prey<(int)ysize_; prey++) {
						//if (prey == yp) continue;
						int fid = edgeid + prey*ysize_+yp;
						tagger_expected[fid] = logsumexp(tagger_expected[fid],
								alpha[node->start][prey] + node->cost + yycost[prey][yp] + beta[i][yp]);
						//cout.precision(15);
						//cout<<"edge: "<<edgeid + prey*ysize_+yp<<" "<<prey<<" "<< yp <<" "<< tagger_expected[fid]<<endl;
					}
				}
			}
		}
	}

	//cout<<"before added expected"<<endl;
	for (map<int, double>::iterator it = tagger_expected.begin(); it != tagger_expected.end(); ++it) {
		//double p = std::exp(it->second - Z_);
		//cout.precision(15);
		//cout<<it->first<<" "<<it->second<<" "<<p<<endl;
		expected[it->first] += std::exp(it->second - Z_);
	}
*/
	//debug
	/*cout<<"expected: "<<endl;
	for (int i = 0; i <feature_index_->size(); ++i) {
		cout.precision(15);
		cout<<expected[i]<<" ";
	}
	cout<<endl;*/
  }

void SemiTagger::calcProb() {
	 int nony = feature_index_->yID("O");
	 for (int i = 0; i < x_.size(); ++i) {//calculate alpha ending at position i with label j
		for(int ns=0; ns<numSegmentsEndAt(i); ns++) {
			double sum = 0.0;
			for (size_t k = 0; k < node_[i][ns].size(); ++k) {
				SegmentNode* node = node_[i][ns][k];
				int yp = node->y;
				double c = 0.0;
				if (node->start-1 < 0) {
					c = node->cost + beta[i][yp];
				} else {
					for(int prey=0; prey<(int)ysize_; prey++) {
						c = logsumexp(c, alpha[node->start][prey] + getTransitionCost(prey,yp), prey==0);
					}
					c += node->cost + beta[i][yp];
				}
				//if (start != end && yp == nony) node->prob = 0.0;
				//else node->prob = std::exp(c - Z_);
				node->prob = std::exp(c - Z_);
				//node->prob = std::exp(c);
				//node->prob = node->cost;
				//node->prob = c-Z_;
				//node->prob = c;
				sum += node->prob;
			}

			for (size_t yp = 0; yp < node_[i][ns].size(); ++yp) {
				//node_[i][ns][yp]->prob = node_[i][ns][yp]->prob/sum;
			}
		}
	 }
}

void SemiTagger::calcCost(SegmentNode *n) {
	n->cost = 0.0;
	double c = 0;
	for (int *f = n->fvector; *f != -1; ++f) {
		c += feature_index_->get_alpha(*f + n->y);
	}

	double *fv = n->embedding_word_vec;
	for (int i = 0; i < feature_index_->word_vec_size; ++i) {
		double v = *fv;
		c += feature_index_->get_alpha(i * feature_index_->embedding_label_size + n->y) * v;
		fv++;
	}

	n->cost = c;
}
// incorporate precision loss comparing to the true segmentation
void SemiTagger::calcCostwithLoss(SegmentNode *n) {
	calcCost(n);

	n->cost_w_loss = n->cost;

	int nony = feature_index_->lb2i["O"];
	// if predict span contains true span for (pos, neg, ...), then no loss
	if (n->y != nony) {
		int correct = 0;
		int non = 0;
		for (int i = n->start; i <= n->end; ++i) {
			if (answer_[i] == n->y) correct++;
			if (answer_[i] == nony) non++;
		}
		int diff = n->end-n->start+1 - correct;
		if (non == diff && diff >=0 && diff <= 3) return;
	}

	double prec_error = 0.0;
	double rec_error = 0.0;
	// boundary loss
	if (n->y == nony) {
		//rec error, n->y = "O", answer = "P"
		for (int i = n->start; i <= n->end; ++i) {
			if (answer_[i] != n->y) {
				int seg_index = token_to_truesegmentation[i];
				if (true_segmentation[seg_index].label != "O") {
					rec_error += (double)1/true_segmentation[seg_index].len();
				}
			}
		}
	} else {
		// prec error n->y = "P", answer = "O"
		int err = 0;
		for (int i = n->start; i <= n->end; ++i) {
			if (answer_[i] != n->y) {
				err++;
			}
		}
		prec_error += (double)err/(n->end-n->start+1);
	}

	n->cost_w_loss += prec_error+rec_error;
}

void SemiTagger::calcBetawithLoss() { //alpha[i][y] = [,y'][,i:y]
	int sent_len = static_cast<int>(x_.size());
	fill(beta_w_loss[sent_len-1].begin(), beta_w_loss[sent_len-1].end(), 0.0);
	for (int i = sent_len - 2; i >= 0;  --i) {
		fill(beta_w_loss[i].begin(), beta_w_loss[i].end(), LOG0);
		// beta[i][y]: marginal ending at i with label y
		for (int yp = 0; yp < ysize_; ++yp) {
			for(int ns=0; ns<numSegmentsStartAt(i+1); ns++) {
				//[,i:y][i+1,:nexty]
				for (size_t k = 0; k < node_beta_[i+1][ns].size(); ++k) {
					int nexty = node_beta_[i+1][ns][k]->y;
					int end = node_beta_[i+1][ns][k]->end;
					beta_w_loss[i][yp] = logsumexp(beta_w_loss[i][yp],
							getTransitionCost(yp,nexty) + node_beta_[i+1][ns][k]->cost_w_loss + beta_w_loss[end][nexty]);
				}
			}
		}
	}
}

void SemiTagger::calcAlphawithLoss() { //alpha[i][y] = [,y'][,i:y]
	int sent_len = static_cast<int>(x_.size());
    fill(alpha_w_loss[0].begin(), alpha_w_loss[0].end(), 0);

    // alpha[i][y] : marginal of segments ending at i with label y
	for (int i = 0; i < sent_len; ++i) {
		fill(alpha_w_loss[i+1].begin(), alpha_w_loss[i+1].end(), LOG0);
		for(int ns=0; ns<numSegmentsEndAt(i); ns++) {
			//[,i-d:][i-d+1,i:y]
			for (size_t k = 0; k < node_[i][ns].size(); ++k) {
				SegmentNode* node = node_[i][ns][k];
				int py = node->y;
				double c = 0.0;
				if (node->start-1 >= 0) {
					for(int prey=0; prey<ysize_; prey++) {
						c = logsumexp(c, alpha_w_loss[node->start][prey]+getTransitionCost(prey,py), prey==0);
					}
				}
				c += node->cost_w_loss;
				alpha_w_loss[i+1][py] = logsumexp(alpha_w_loss[i+1][py], c);
			}
		}
	}

	// debug
#ifdef DEBUG_
	cout<<"alpha"<<endl;
	for (int i = 0; i < sent_len+1;  ++i) {
		for (int y = 0; y < ysize_; ++y) {
			cout<<alpha[i][y]<<" ";
		}
		cout<<endl;
	}
	cout<<endl;
#endif
  }

  void SemiTagger::calcAlpha() { //alpha[i][y] = [,y'][,i:y]
	int sent_len = static_cast<int>(x_.size());
    fill(alpha[0].begin(), alpha[0].end(), 0);

    // alpha[i][y] : marginal of segments ending at i with label y
	for (int i = 0; i < sent_len; ++i) {
		fill(alpha[i+1].begin(), alpha[i+1].end(), LOG0);
		for(int ns=0; ns<numSegmentsEndAt(i); ns++) {
			//[,i-d:][i-d+1,i:y]
			for (size_t k = 0; k < node_[i][ns].size(); ++k) {
				SegmentNode* node = node_[i][ns][k];
				int py = node->y;
				double c = 0.0;
				if (node->start-1 >= 0) {
					for(int prey=0; prey<ysize_; prey++) {
						c = logsumexp(c, alpha[node->start][prey]+getTransitionCost(prey,py), prey==0);
					}
				}
				c += node->cost;
				alpha[i+1][py] = logsumexp(alpha[i+1][py], c);
			}
		}
	}

	// debug
#ifdef DEBUG_
	cout<<"alpha"<<endl;
	for (int i = 0; i < sent_len+1;  ++i) {
		for (int y = 0; y < ysize_; ++y) {
			cout<<alpha[i][y]<<" ";
		}
		cout<<endl;
	}
	cout<<endl;
#endif
  }

  void SemiTagger::calcBeta() {//beta[i][y] = [i:y,][i+d,:nexty]
	int sent_len = static_cast<int>(x_.size());
	fill(beta[sent_len-1].begin(), beta[sent_len-1].end(), 0.0);
	for (int i = sent_len - 2; i >= 0;  --i) {
		fill(beta[i].begin(), beta[i].end(), LOG0);
		// beta[i][y]: marginal ending at i with label y
		for (int yp = 0; yp < ysize_; ++yp) {
			for(int ns=0; ns<numSegmentsStartAt(i+1); ns++) {
				//[,i:y][i+1,:nexty]
				for (size_t k = 0; k < node_beta_[i+1][ns].size(); ++k) {
					int nexty = node_beta_[i+1][ns][k]->y;
					int end = node_beta_[i+1][ns][k]->end;
					beta[i][yp] = logsumexp(beta[i][yp],
							getTransitionCost(yp,nexty) +node_beta_[i+1][ns][k]->cost+beta[end][nexty]);
				}
			}
		}
	}

	// debug
#ifdef DEBUG_
	cout<<"beta"<<endl;
	for (int i = 0; i < sent_len;  ++i) {
		for (int y = 0; y < ysize_; ++y) {
			cout.precision(20);
			cout<<beta[i][y]<<" ";
		}
		cout<<endl;
	}
	cout<<endl;
#endif
  }

  void SemiTagger::ComputeCostWithLoss(bool loss_aware) {
 	//cout<<"compute cost"<<endl;
  	//calculate sum of feature values for each node
      for (size_t i = 0; i < x_.size(); ++i) {
  		for(int ns = 0; ns<numSegmentsEndAt(i); ns++) {
  		  for (size_t j = 0; j < node_[i][ns].size(); ++j) {
  			  if (loss_aware) {
  				  calcCostwithLoss(node_[i][ns][j]);
  			  } else {
  				  calcCost(node_[i][ns][j]);
  				  node_[i][ns][j]->cost_w_loss = node_[i][ns][j]->cost;
  			  }
  			  //cout<<"node "<<node_[i][ns][j]->start<<","<<node_[i][ns][j]->end<<" "<<node_[i][ns][j]->cost<<endl;
  		  }
  		}
      }

      // compute transition cost
     int fid = feature_index_->edgeid_;
     if (fid != -1) {
		for(int y=0; y<ysize_; y++) {
			for(int prey=0; prey<ysize_; prey++) {
				//if (prey == y) {
				//	yycost[prey][y] = 0.0;
				//} else {
					double cost = calcCost(fid, prey, y);
					yycost[prey][y] = cost;
					//cout<<"edge "<<prey<<" "<<y<<" "<<cost<<endl;
				//}
			}
		}
	}
  }

  void SemiTagger::ComputeCost() {
	//cout<<"compute cost"<<endl;
 	//calculate sum of feature values for each node
     for (size_t i = 0; i < x_.size(); ++i) {
 		for(int ns = 0; ns<numSegmentsEndAt(i); ns++) {
 		  for (size_t j = 0; j < node_[i][ns].size(); ++j) {
 			  calcCost(node_[i][ns][j]);
 			  //cout<<"node "<<node_[i][ns][j]->start<<","<<node_[i][ns][j]->end<<" "<<node_[i][ns][j]->cost<<endl;
 		  }
 		}
     }

     // compute transition cost
    int fid = feature_index_->edgeid_;
    if (fid != -1) {
		for(int y=0; y<ysize_; y++) {
			for(int prey=0; prey<ysize_; prey++) {
				//if (prey == y) {
				//	yycost[prey][y] = 0.0;
				//} else {
					double cost = calcCost(fid, prey, y);
					yycost[prey][y] = cost;
					//cout<<"edge "<<prey<<" "<<y<<" "<<cost<<endl;
				//}
			}
		}
    }
  }

   void SemiTagger::forwardbackward() {
     if (x_.empty()) return;

     // debug
     /*cout<<"lambda: "<<endl;
     for (int i = 0; i < feature_index_->size(); ++i) {
    	 	 cout<<feature_index_->get_alpha(i)<<" ";
     }
     cout<<endl;*/

    calcBeta();
 	calcAlpha();

 	 Z_ = 0.0;
 	 for (size_t j = 0; j < ysize_; ++j)
 		 Z_ = logsumexp(Z_, alpha[x_.size()][j], j == 0);

     return;
   }

   double SemiTagger::SpanLoss(int start, int end, int y, int y_hat) {
   	double c = 0;
   	if (y == y_hat) c = 0;
   	else {
   		if (start == end) {
   			c = 1;
   		} else {
   			c = 0.1;
   		}
   	}
   	return c;
   }

   void SemiTagger::viterbiwithLoss() {
     int xsize_ = (int)x_.size();

     // compute cost as negative expected loss
     calcProb();

     for (size_t i = 0; i < xsize_; ++i) {
		for(int ns=0; ns<numSegmentsEndAt(i); ns++) {
			for (size_t k = 0; k < node_[i][ns].size(); ++k) {
				SegmentNode *node = node_[i][ns][k];
				int py = node->y;
				//update cost
				double exploss = 0.0;
				//cout<<"pred "<<endl;
				for (int k_hat = 0; k_hat < node_[i][ns].size(); ++k_hat) {
					int y_hat = node_[i][ns][k_hat]->y;
					exploss += SpanLoss(node->start, node->end, py, y_hat) * node_[i][ns][k_hat]->prob;
					//cout<<node_[i][ns][k_hat]->prob<<" ";
				}
				//cout<<endl;
				node->cost = -exploss;
				//cout<<"node "<<node->start<<" "<<node->end<<" "<<node->y<<" "<<exploss<<endl;
			}
		}
	}
    //cout<<endl;

     int nony = feature_index_->yID("O");
     for (size_t i = 0; i < xsize_; ++i) {
    	 	 for(int ns=0; ns<numSegmentsEndAt(i); ns++) {
    	 		 for (size_t k = 0; k < node_[i][ns].size(); ++k) {
    	 			 SegmentNode *node = node_[i][ns][k];
    	 			 int py = node->y;

				// best prev
				double bestc = -1e37;
				SegmentNode *best = 0;
				if (node->start > 0) {
					for (int nt = 0; nt < numSegmentsEndAt(node->start-1); ++nt) {
						for (int k = 0; k < node_[node->start-1][nt].size(); ++k) {
							double cost = node_[node->start-1][nt][k]->bestCost + node->cost;
							if (cost > bestc) {
								bestc = cost;
								best  = node_[node->start-1][nt][k];
							}
						}
					}
				}
				node->prev     = best;
				node->bestCost = best ? bestc : node->cost;
			}
		}
	}

   	double bestc = -1e37;
   	SegmentNode *best = 0;
   	size_t s = xsize_-1;
   	for (int ns = 0; ns < numSegmentsEndAt(s); ++ns) {
   		for (size_t j = 0; j < node_[s][ns].size(); ++j) {
   			 if (bestc < node_[s][ns][j]->bestCost) {
   			   best  = node_[s][ns][j];
   			   bestc = node_[s][ns][j]->bestCost;
   			 }
   		}
   	}

   	for (SegmentNode *n = best; n; n = n->prev) {
   		for (int i = n->start; i <= n->end; ++i) {
   	         result_[i] = n->y;
   		}
   	}

   	cost_ = -best->bestCost;

   }

   void SemiTagger::viterbi() {
  	int xsize_ = (int)x_.size();
  	int nony = feature_index_->yID("O");

  	for (size_t i = 0; i < xsize_; ++i) {
  		for(int ns=0; ns<numSegmentsEndAt(i); ns++) {
  			for (size_t k = 0; k < node_[i][ns].size(); ++k) {
  				SegmentNode *node = node_[i][ns][k];
  				int py = node->y;
				// best prev
				double bestc = -1e37;
				SegmentNode *best = 0;
				if (node->start > 0) {
					for (int nt = 0; nt < numSegmentsEndAt(node->start-1); ++nt) {
						for (int k = 0; k < node_[node->start-1][nt].size(); ++k) {
							int prey = node_[node->start-1][nt][k]->y;
							double cost = node_[node->start-1][nt][k]->bestCost
								  + getTransitionCost(prey,py) + node->cost;
							if (cost > bestc) {
								bestc = cost;
								best  = node_[node->start-1][nt][k];
							}
						}
					}
				}
				node->prev     = best;
				node->bestCost = best ? bestc : node->cost;
			}
		}
	}

	double bestc = -1e37;
	SegmentNode *best = 0;
	size_t s = xsize_-1;
	for (int ns = 0; ns < numSegmentsEndAt(s); ++ns) {
		for (size_t j = 0; j < node_[s][ns].size(); ++j) {
			 if (bestc < node_[s][ns][j]->bestCost) {
			   best  = node_[s][ns][j];
			   bestc = node_[s][ns][j]->bestCost;
			 }
		}
	}

	for (SegmentNode *n = best; n; n = n->prev) {
		for (int i = n->start; i <= n->end; ++i) {
	         result_[i] = n->y;
		}
	}

	cost_ = -best->bestCost;

   }

   bool SemiTagger::Inference() {
         if (x_.empty()) return true;

         feature_index_->buildSemiCRFTagger(this);
         ComputeCost();

         forwardbackward();

         viterbi();

         if (nbest_) {
     		initNbest();
     	 }

         return true;
       }

   bool SemiTagger::InferenceWithLoss() {
	if (x_.empty()) return true;

	feature_index_->buildSemiCRFTagger(this);
	ComputeCost();

	forwardbackward();

	viterbiwithLoss();

	return true;
  }

   bool SemiTagger::initNbest() {
      if (!agenda_.get()) {
        agenda_.reset(new std::priority_queue <QueueSegElement*,
                      std::vector<QueueSegElement *>, QueueSegElementComp>);
        nbest_freelist_.reset(new FreeList <QueueSegElement>(128));
      }

      nbest_freelist_->free();
      while (!agenda_->empty()) agenda_->pop();   // make empty

      size_t k = x_.size()-1;
      for (int ns = 0; ns < numSegmentsEndAt(k); ++ns) {
		  for (size_t i = 0; i < ysize_; ++i) {
			QueueSegElement *eos = nbest_freelist_->alloc();
			eos->node = node_[k][ns][i];
			eos->fx = -node_[k][ns][i]->bestCost;
			eos->gx = -node_[k][ns][i]->cost;
			eos->next = 0;
			agenda_->push(eos);
		  }
      }

      return true;
    }

    bool SemiTagger::next() {
      while (!agenda_->empty()) {
        QueueSegElement *top = agenda_->top();
        agenda_->pop();
        SegmentNode *rnode = top->node;

        if (rnode->start == 0) {
          for (QueueSegElement *n = top; n; n = n->next) {
        	  	  for(int k=n->node->start; k<=n->node->end; k++) {
        	  		  result_[k] = n->node->y;
        	  	  }
          }
          cost_ = top->gx;
          return true;
        }

        for(int ns=0; ns<numSegmentsEndAt(rnode->start-1); ns++) {
			for (size_t j = 0; j < ysize_; ++j) {
				SegmentNode* node = node_[rnode->start-1][ns][j];

				QueueSegElement *n =nbest_freelist_->alloc();
				n->node = node;
				n->gx   = -node->cost - getTransitionCost(j,rnode->y) + top->gx;
				n->fx   = -node->bestCost -getTransitionCost(j,rnode->y) +  top->gx;
				//          |              h(x)                 |  |  g(x)  |
				n->next = top;
				agenda_->push(n);
			}
        }
      }

      return false;
    }

    bool isRepeated(vector<vector<unsigned short int> > &nbest_results,
    		vector<unsigned short int> &results) {
    		for (int i = 0; i < nbest_results.size(); ++i) {
    			bool same = true;
    			for (int j = 0; j < results.size(); ++j) {
    				if (results[j] != nbest_results[i][j]) {
    					same = false;
    					break;
    				}
    			}
    			if (same) {
    				return true;
    			}
    		}
    		return false;
    }

    string SemiTagger::toString() {
		string str = "";
		if(nbest_) {
			// output each token has nbest tags
			vector<string> results(x_.size());
			fill(results.begin(), results.end(), "");

			vector<double> probs;
			int n = 0;
			int c = 0;
			vector<vector<unsigned short int> > nbest_results;
			while(n < nbest_) {
				if (!next()) {
					break;
				}

				double p = prob();
				if(!isRepeated(nbest_results, result_)) {
					probs.push_back(p);
					for (size_t i = 0; i < x_.size(); ++i) {
						string str = yname(y(i));
						results[i] += str + "\t";
					}
					nbest_results.push_back(result_);
					n++;
				}
				c++;
				if (c == 1000) break;
			}
			vector<vector<string> > nbest_results_str(nbest_results.size());
			for (int k = 0; k < nbest_results.size(); ++k) {
				nbest_results_str[k].resize(nbest_results[k].size(), "O");
				for (size_t i = 0; i < nbest_results[k].size(); ++i) {
					string label = feature_index_->ystr(nbest_results[k][i]);

					if (label == "O") continue;
					nbest_results_str[k][i] = label;
				}

				int i = 0;
				while (i < nbest_results_str[k].size()) {
					if (nbest_results_str[k][i] != "O") {
						string label = nbest_results_str[k][i];
						nbest_results_str[k][i] = "B_"+nbest_results_str[k][i];
						i++;
						while (i < nbest_results_str[k].size()
								&& nbest_results_str[k][i] == label) i++;
					} else {
						i++;
					}
				}
			}

			vector<string> answers(sent->size(), "O");
			for (int i = 0; i < sent->entities.size(); ++i) {
				int start = sent->entities[i].start;
				int end = sent->entities[i].end;
				string label = sent->entities[i].label;
				answers[start] = "B_"+label;
				for (int j = start+1; j <= end; ++j) {
					answers[j] = label;
				}
			}

			for (size_t i = 0; i < x_.size(); ++i) {
				str += x_[i][0];
				str += "\t";
				str += answers[i];
				for (int j = 0; j <nbest_results_str.size(); ++j) {
					str += "\t" + nbest_results_str[j][i];
				}
				str += "\n";
			}
			str += "\n";
			for (int i = 0; i < probs.size(); ++i) {
				str += Utils::int2string(i)+":  "+Utils::double2string(probs[i])+"\n";
			}
			str += "\n";
		} else {
			for (size_t i = 0; i < x_.size(); ++i) {
				str += x_[i][0];
				str += "\t";
				str += yname(answer_[i]);
				str += "\t";
				str += yname(result_[i]);
				str += "\n";
			}
			str += "\n";
		}
		return str;
  }

   int SemiTagger::eval() {
     int err = 0;
     for (size_t i = 0; i < x_.size(); ++i)
       if (answer_[i] != result_[i]) { ++err; }
     return err;
   }

   bool SemiTagger::clear() {
     //if (mode_ == TEST && feature_index_) feature_index_->clear();
 	//if (mode_ == TEST && feature_index_) feature_index_->clear();

    x_.clear();

 	node_.clear();
 	node_beta_.clear();

 	alpha.clear();
 	beta.clear();

 	alpha_w_loss.clear();

 	segmentsEnd.clear();
 	segmentsStart.clear();

     answer_.clear();
     result_.clear();

     Z_ = cost_ = 0.0;
     return true;
   }

  bool SemiTagger::shrink() {
/*     std::vector<std::vector<string> >(x_).swap(x_);
     std::vector<std::vector<vector<SegmentNode *> > >(node_).swap(node_);
     std::vector<unsigned short int>(answer_).swap(answer_);
     std::vector<unsigned short int>(result_).swap(result_);

     std::vector<Entity>(truesegments).swap(truesegments);
     vector<string>(tags).swap(tags);
     vector<string>(resulttags).swap(resulttags);
     vector<vector<int> >(segmentsEnd).swap(segmentsEnd); //segments end at i
     vector<vector<int> >(segmentsStart).swap(segmentsStart); //segments start at i
*/
     return true;
   }
  bool SemiTagger::open(FeatureIndex *f) {
    mode_ = LEARN;
    feature_index_ = f;
    ysize_ = feature_index_->ysize();
    return true;
  }
  bool SemiTagger::test_open(FeatureIndex *f) {
    mode_ = TEST;
    feature_index_ = f;
    ysize_ = feature_index_->ysize();
    return true;
  }

  void SemiTagger::close() {
    if (mode_ == TEST) {
    }
  }

  bool SemiTagger::add2(size_t size, const char **column, bool copy) {
	/*
    if ((mode_ == LEARN && size < xsize + 1) ||
        (mode_ == TEST  && size < xsize)) {
      CHECK_FALSE(false) << "# x is small: size="
        << size << " xsize=" << xsize;
    }
	*/

    size_t s = x_.size() + 1;
    x_.resize(s);

	segmentsEnd.resize(s);
	segmentsStart.resize(s);

    answer_.resize(s);
    result_.resize(s);

    s = x_.size() - 1;

    if (copy) {
      for (size_t k = 0; k < size; ++k)
        x_[s].push_back(feature_index_->strdup(column[k]));
    } else {
      for (size_t k = 0; k < size; ++k)
        x_[s].push_back(column[k]);
    }

    result_[s] = answer_[s] = 0;  // dummy
    //if (mode_ == LEARN) {
      size_t r = ysize_;
      for (size_t k = 0; k < ysize_; ++k)
      {
    	  string label = column[size-1];

    	  if(label.substr(0,2) == "B_")
    		  label = label.substr(2);
        if (std::strcmp(yname(k), label.c_str()) == 0)
        {
          r = k;
          answer_[s] = r;
          break;
        }
      }

      CHECK_FALSE(r != ysize_) << "cannot find answer: " << column[size-1];

    //}

    return true;
  }

  bool SemiTagger::add(size_t size, const char **column) {
    return add2(size, column, true);
  }

  bool SemiTagger::add(const char* line) {
    const char* column[8192];
    char *p = feature_index_->strdup(line);
    size_t size = tokenize2(p, "\t ", column, sizeof(column));
    if (!add2(size, column, false)) return false;
    return true;
  }

  bool SemiTagger::read(std::istream *is) {
    char line[8192];
    clear();

    for (;;) {
      if (!is->getline(line, sizeof(line))) {
        is->clear(std::ios::eofbit|std::ios::badbit);
        return true;
      }
      if (line[0] == '\0' || line[0] == ' ' || line[0] == '\t') break;
      if (!add(line)) return false;
    }

    return true;
  }

  void SemiTagger::sortSegEnds() {
	  for(int i=0; i<segmentsEnd.size(); i++) {
		  sort(segmentsEnd[i].begin(), segmentsEnd[i].end(), greater<int>());
	  }
  }

  void SemiTagger::sortSegStarts() {
	  for(int i=0; i<segmentsStart.size(); i++) {
		  sort(segmentsStart[i].begin(), segmentsStart[i].end());
	  }
  }

void SemiTagger::genSegmentUnitCandidates()
{
	map<int, set<int> > start_maps;
	// include each word as a unit
	for (int i = 0; i <size(); ++i) {
		set<int> p;
		p.insert(i);
		start_maps[i] = p;
	}

	for (int i = 0; i < span_units.size(); ++i) {
		int s = i;
		string label = span_units[i].label;
		int j = i+1;
		// find segments belonging to the segment group;
		while (j < span_units.size()) {
			// looking at the gold standard label
			if (span_units[j].label != label) {
				break;
			}
			j++;
		}
		int t = j-1;

		// enumerate spans [s,t]
		int start_pos = span_units[s].start;
		for (int k = s; k <= t; ++k) {
			int end_pos = span_units[k].end;
			start_maps[start_pos].insert(end_pos);
		}
	}

	for (map<int, set<int> >::iterator it = start_maps.begin();
			it != start_maps.end(); ++it) {
		for (set<int>::iterator sit = it->second.begin();
				sit != it->second.end(); ++sit) {
			int start_pos = it->first;
			int end_pos = *sit;
			segmentsEnd[end_pos].push_back(start_pos);
			segmentsStart[start_pos].push_back(end_pos);
		}
	}

	sortSegEnds();
	sortSegStarts();
}

  void SemiTagger::genSegmentCandidates(int max_length)
  {
	 for (size_t i = 0; i < size(); ++i) {
		 for (size_t l = 1; l <= max_length; ++l) {
			int start_pos = i;
			int end_pos = i+l-1;
			if (end_pos < size()) {
				segmentsEnd[end_pos].push_back(start_pos);
				segmentsStart[start_pos].push_back(end_pos);
			}
		}
	 }

	 sortSegEnds();
	 sortSegStarts();
  }


} /* namespace CRFPP */
