/*
 * semi_tagger.h
 *
 *  Created on: Feb 15, 2014
 *      Author: bishan
 */

#ifndef SEMI_TAGGER_H_
#define SEMI_TAGGER_H_
#include <iostream>
#include <vector>
#include <algorithm>
#include <map>
#include "scoped_ptr.h"
#include "node.h"
#include "tagger.h"
#include "feature_index.h"
#include "../Parsing/Utils.h"
#include "../Sentence.h"

#define MAX_LENGTH 5

namespace CRFPP {

class SemiTagger : public Tagger{
public:
	explicit SemiTagger(): mode_(TEST), vlevel_(0), nbest_(0),
						  ysize_(0), Z_(0),
						  inference_type(0), costC(1) {
		sent = NULL;
		feature_id_ = 0;
		feature_value_id_ = 0;
		thread_id_ = 0;
		feature_index_ = 0;
	}
    ~SemiTagger() { close(); }
private:
   enum { TEST, LEARN };

   unsigned int mode_   : 2;
   unsigned int vlevel_ : 3;

   int           inference_type;
   double        costC;

   double                            cost_;

   std::vector<std::vector <string> > x_;

   whatlog what_;
   string_buffer os_;

   scoped_ptr<std::priority_queue <QueueSegElement*, std::vector <QueueSegElement *>,
     QueueSegElementComp> > agenda_;
   scoped_ptr<FreeList <QueueSegElement> > nbest_freelist_;
   QueueSegElement* bestQueElem;
 public:
   unsigned int nbest_  : 11;
   size_t    ysize_;
   double                            Z_;

	vector<int>  token_to_entities;

   std::vector<std::vector<vector<SegmentNode *> > > node_;
   std::vector<std::vector<vector<SegmentNode *> > > node_beta_;

	vector<vector<double> > alpha; //|x| * |y|
	vector<vector<double> > beta; //|x| * |y|

	vector<vector<double> > alpha_w_loss; //|x| * |y|
	vector<vector<double> > beta_w_loss; //|x| * |y|

	vector<vector<double> >  yycost; //ysize_*ysize_

	vector<vector<int> > segmentsEnd; //segments end at i
	vector<vector<int> > segmentsStart; //segments start at i

	vector<Span> span_units;
	vector<Span> true_segmentation;
	vector<int>  token_to_truesegmentation;

public:
	void  calcExpectationWithLoss(double *expected);
	void  calcBetawithLoss();
	void  ComputeCostWithLoss(bool loss_aware);

	bool  IsInSameSegment(int i, int j);
	void   BuildUnitBasedTrueSegmentation();
	void   BuildTrueSegmentation(int max_len);
	bool IsTrueSegment(int start, int end);

	double getTransitionCost(int prey, int y) {
		return yycost[prey][y];
	}

	bool InferenceWithLoss();
	double SpanLoss(int start, int end, int y, int y_hat);
	void calcProb();
	void viterbiwithLoss();
	string toString();

	bool SentenceToSemiTagger(Sentence *psent,
			CRFPP::FeatureIndex *feature_index, bool train, bool useparse=false);

	void calcCost(SegmentNode *n);

	void calcExpectation(double *expected);

	void calcAlpha();
	void calcBeta();

	void calcAlphawithLoss();

	// incorporate precision loss comparing to the true segmentation
	void calcCostwithLoss(SegmentNode *n);

	double calcCost(int fid, int prey, int y) {
	   double c = feature_index_->get_alpha(fid + prey * ysize_ + y);     \
	   return c;
	}

	  int get_lb2i(string str);
	  string get_i2lb(int i);

	  void setNbest(int n) {nbest_ = n;}
	  int ysize() {return ysize_;}

	  void initializeLattice();

		void forwardbackward();
		void viterbi();
		void ComputeCost();
		bool initNbest();
		bool add2(size_t, const char **, bool);

	  // using span units
	  void genSegmentUnitCandidates();
	  void genSegmentCandidates(int max_length);

	  void sortSegEnds();
	  void sortSegStarts();

	  void initTagger();

	  double       prob() const { return std::exp(- cost_ - Z_); }
	  double       toprob(SegmentNode *n);
	  double       toyyprob(SegmentNode* n, int y);
	  double       prob(size_t i, size_t j, size_t k) {
		  return toprob(node_[i][j][k]);
	  }


       SegmentNode*  node(size_t i, size_t j, size_t k) const { return node_[i][j][k]; }
 	  void   set_node(SegmentNode *n, size_t i, size_t j, size_t k) { node_[i][j][k] = n; }

 	  SegmentNode*  node_beta(size_t i, size_t j, size_t k) const { return node_beta_[i][j][k]; }
 	  void   set_node_beta(SegmentNode *n, size_t i, size_t j, size_t k) { node_beta_[i][j][k] = n; }


	    int nextSegEnd_StartAt(int i, int j)
		{
			if(i < 0 || i>= (int) segmentsStart.size())
				return -1;
			if(j < 0 || j >= (int)segmentsStart[i].size())
				return -1;
			return segmentsStart[i][j];
		}
		int nextSegStart_EndAt(int i, int j)
		{
			if(i < 0 || i >= (int) segmentsEnd.size())
				return -1;
			if(j < 0 || j >= (int)segmentsEnd[i].size())
				return -1;
			return segmentsEnd[i][j];
		}
		int numSegmentsStartAt(int i)
		{
			if(i < 0 || i>= (int) segmentsStart.size())
				return 0;
			return (int)segmentsStart[i].size();
		}
		int numSegmentsEndAt(int i)
		{
			if(i < 0 || i>= (int) segmentsEnd.size())
				return 0;
			return (int)segmentsEnd[i].size();
		}

		SegmentNode* GetSegmentNode(int start, int end, int py) {
			for (int i = 0; i < node_[end].size(); ++i) {
				for (int j = 0; j < node_[end][i].size(); ++j) {
					if (node_[end][i][j]->start == start && node_[end][i][j]->y == py) {
						return node_[end][i][j];
					}
				}
			}
			return NULL;
		}

		void getNodebySegmentEnd(int start, int end, vector<SegmentNode*>& v)
		{
			for(int i =0; i<(int)segmentsEnd[end].size(); i++) {
				if(segmentsEnd[end][i] == start) {
					v.resize(node_[end][i].size());
					for(int j=0; j<(int)node_[end][i].size(); j++) {
						v[j] = node_[end][i][j];
					}
					return;
				}
			}
			cout<<"couldn't find true segment!!"<<endl;
		}

		void getNodebySegmentStart(int start, int end, vector<SegmentNode*>& v)
		{
			for(int i =0; i<(int)segmentsStart[start].size(); i++) {
				if(segmentsStart[start][i] == end) {
					v.resize(node_beta_[start][i].size());
					for(int j=0; j<(int)node_beta_[start][i].size(); j++)
						v[j] = node_beta_[start][i][j];
					return;
				}
			}
		}


   int          eval();
   double       gradient(double *);

   bool         shrink();
   bool         parse_stream(std::istream *, ifstream& segis, ofstream& os);
   bool         read(std::istream *);

   bool         open(FeatureIndex *);
   bool         test_open(FeatureIndex *);

   bool         open(const char*);
   bool         open(int, char **);
   void         close();
   bool         add(size_t, const char **);
   bool         add(const char*);
   size_t       size() const { return x_.size(); }

   size_t       dsize() const { return feature_index_->size(); }
   const float *weight_vector() const { return feature_index_->alpha_float(); }
   bool         empty() const { return x_.empty(); }
   size_t ysize() const { return ysize_; }
   double cost() const { return cost_; }
   double Z() const { return Z_; }

   size_t answer(size_t i) const { return answer_[i]; }
   size_t result(size_t i) const { return result_[i]; }
   size_t y(size_t i)  const    { return result_[i]; }
   const char* yname(size_t i) const    { return feature_index_->ystr(i); }
   const char* y2(size_t i) const      { return yname(result_[i]); }

   bool Inference();
   bool clear();
   bool next();

   unsigned int vlevel() const { return vlevel_; }

   float cost_factor() const {
     return feature_index_->cost_factor();
   }

   size_t nbest() const { return nbest_; }

   void set_vlevel(unsigned int vlevel) {
     vlevel_ = vlevel;
   }

   void set_cost_factor(float cost_factor) {
     if (cost_factor > 0)
       feature_index_->set_cost_factor(cost_factor);
   }

   void set_nbest(size_t nbest) {
     nbest_ = nbest;
   }

   const char* what() { return what_.str(); }

};

} /* namespace CRFPP */
#endif /* SEMI_TAGGER_H_ */
