//
//  CRF++ -- Yet Another CRF toolkit
//
//  $Id: tagger.cpp 1601 2007-03-31 09:47:18Z taku $;
//
//  Copyright(C) 2005-2007 Taku Kudo <taku@chasen.org>
//
#include <iostream>
#include <vector>
#include <iterator>
#include <cmath>
#include <string>
#include <sstream>
#include "common.h"
#include "tagger.h"
#include <stdio.h>
#include <stdlib.h>

//#define DEBUG_

std::string errorStr;

namespace CRFPP {

bool TaggerImpl::open(FeatureIndex *f) {
    mode_ = LEARN;
    feature_index_ = f;
    ysize_ = feature_index_->ysize();
    zero_order = feature_index_->zero_order;

    return true;
  }

  bool TaggerImpl::testopen(string modelfile)
  {
	    close();
	  	feature_index_ = new DecoderFeatureIndex();
	  	CHECK_CLOSE_FALSE(feature_index_->open(modelfile.c_str()))
	  	      << feature_index_->what();
	   ysize_ = feature_index_->ysize();

	   return true;
  }

  void TaggerImpl::close() {
    if (mode_ == TEST) {
      //delete feature_index_;
      //feature_index_ = 0;
    }
  }

  bool TaggerImpl::add2(size_t size, const char **column, bool copy) {
    /*size_t xsize = feature_index_->xsize();

    if ((mode_ == LEARN && size < xsize + 1) ||
        (mode_ == TEST  && size < xsize)) {
      CHECK_FALSE(false) << "# x is small: size="
        << size << " xsize=" << xsize;
    }*/

    size_t s = x_.size() + 1;
    x_.resize(s);
    node_.resize(s);
    answer_.resize(s);
    result_.resize(s);

    s = x_.size() - 1;

    result_[s] = answer_[s] = 0;  // dummy
    //if (mode_ == LEARN) {
      size_t r = ysize_;
      for (size_t k = 0; k < ysize_; ++k)
        if (std::strcmp(yname(k), column[size-1]) == 0)
          r = k;

      if (r == ysize_) {
    	     cout<<"cannot find answer: " << column[size-1]<<endl;
    	     return false;
      }
      answer_[s] = r;
    //}

    node_[s].resize(ysize_);

    return true;
  }

  bool TaggerImpl::add(const char* line) {
    const char* column[8192];
    char *p = feature_index_->strdup(line);
    size_t size = tokenize2(p, "\t ", column, sizeof(column));
    if (!add2(size, column, false)) {
    		return false;
    }
    return true;
  }

  bool TaggerImpl::read(std::istream *is) {
    char line[8192];
    clear();

    for (;;) {
      if (!is->getline(line, sizeof(line))) {
        is->clear(std::ios::eofbit|std::ios::badbit);
        return true;
      }
      if (line[0] == '\0' || line[0] == ' ' || line[0] == '\t') break;
      if (!add(line)) return false;
    }

    return true;
  }

  bool TaggerImpl::readinst(vector<string> instlines) {
    clear();

	for(int i=0; i<instlines.size(); i++) {
		if (!add(instlines[i].c_str())) {
			return false;
		}
		// dummy word
		x_[i].push_back("w");
	}
    return true;
  }

  bool TaggerImpl::shrink(bool train) {
	feature_index_->buildCRFFeatures(this, train);

    std::vector<std::vector<const char *> >(x_).swap(x_);
    std::vector<std::vector<Node *> >(node_).swap(node_);
    std::vector<unsigned short int>(answer_).swap(answer_);
    std::vector<unsigned short int>(result_).swap(result_);

    return true;
  }

  int TaggerImpl::eval() {
    int err = 0;
    for (size_t i = 0; i < x_.size(); ++i)
      if (answer_[i] != result_[i]) { ++err; }
    return err;
  }

  void TaggerImpl::eval(map<pair<int, int>, int> &confusion_matrix) {
      for (size_t i = 0; i < x_.size(); ++i) {
    	  confusion_matrix[make_pair(answer_[i] ,result_[i])] += 1;
      }
    }

  bool TaggerImpl::clear() {
    x_.clear();
    node_.clear();
    answer_.clear();
    result_.clear();

    Z_ = 0.0;

    return true;
  }

  //For inference, recompute cost and constraintcost
  void TaggerImpl::buildLattice() {
    if (x_.empty()) return;

    //clear node structure
    for (size_t i = 0; i < x_.size(); ++i) {
         for (size_t j = 0; j < ysize_; ++j) {
        	 node_[i][j]->zeros();

        	 vector<Path *> &lpath = node_[i][j]->lpath;
			 for (const_Path_iterator it = lpath.begin(); it != lpath.end(); ++it)
			     (*it)->zeros();
         }
       }

    recomputeCost();
  }

  double TaggerImpl::labelMarginal() {
      if (x_.empty()) return 0.0;
      double marginal = 0.0;
      for (int i = 0; i < static_cast<int>(x_.size()); ++i) {
        for (size_t j = 0; j < ysize_; ++j) {
			marginal += node_[i][j]->prob * node_[i][j]->cost;
			const std::vector<Path *> &lpath = node_[i][j]->lpath;
			for (const_Path_iterator it = lpath.begin(); it != lpath.end(); ++it)
				marginal += (*it)->prob * (*it)->cost;
        }
      }
      return marginal;
  }
  double TaggerImpl::gradient(double *expected)
  {
       if (x_.empty()) return 0.0;

       // reallocate the node structure
       feature_index_->buildCRFTagger(this);

       // check node features
/*       cout<<"tagger features: "<<endl;
       for (size_t i = 0; i < x_.size(); ++i) {
    	      cout<<"node "<<i<<": ";
		  for (size_t j = 0; j < ysize_; ++j) {
			  cout<<"\t label "<<j<<": ";
			  for (int *f = node_[i][j]->fvector; *f != -1; ++f) {
				  cout<<feature_index_->feature_vec[*f + node_[i][j]->y]<<" ";
			  }
			  //cout<<endl;
		  }
		  cout<<endl;
       }
       cout<<endl;
*/
       //clear node structure

       /*recomputeCost();
  	   forwardbackward(); //calculate alpha, beta and Z_

		Z_ = 0.0;
		if (!zero_order) {
			for (size_t j = 0; j < ysize_; ++j) {
				Z_ = logsumexp(Z_, node_[x_.size() - 1][j]->alpha, j == 0);
			}
			for (size_t i = 0; i < x_.size(); ++i) {
				for (size_t j = 0; j < ysize_; ++j) {
					node_[i][j]->prob = std::exp(node_[i][j]->alpha + node_[i][j]->beta - Z_);
					//cout<<"node: "<<i<<" "<<j<<" "<<node_[i][j]->prob<<endl;
					for (const_Path_iterator it = node_[i][j]->lpath.begin(); it != node_[i][j]->lpath.end(); ++it) {
						(*it)->prob = std::exp((*it)->lnode->alpha + (*it)->cost + (*it)->rnode->cost + (*it)->rnode->beta - Z_);
						//cout<<"edge: "<<(*it)->lnode->y<<" "<<(*it)->rnode->y<<" "<<(*it)->prob<<endl;
					}
				}
			}
		} else {
			for (size_t i = 0; i < x_.size(); ++i) {
				double sum_x = 0.0;
				for (size_t j = 0; j < ysize_; ++j) {
					sum_x = logsumexp(sum_x, node_[i][j]->alpha, (j == 0));
				}
				Z_ += sum_x;
				for (size_t j = 0; j < ysize_; ++j) {
					node_[i][j]->prob = std::exp(node_[i][j]->alpha - sum_x);
				}
			}
		}
     */

       recomputeCostwithLoss();
  	 // Compute marginals
  	    for (int i = 0; i < static_cast<int>(x_.size()); ++i) {
  	 		 for (size_t j = 0; j < ysize_; ++j) {
  	 			 node_[i][j]->calcAlphaWithLoss();
  	 		 }
  	 	}
  	   	for (int i = static_cast<int>(x_.size() - 1); i >= 0;  --i) {
  	 		 for (size_t j = 0; j < ysize_; ++j) {
  	 			 node_[i][j]->calcBetaWithLoss();
  	 		 }
  	 	}

  		Z_ = 0.0;
  		if (!zero_order) {
  			for (size_t j = 0; j < ysize_; ++j) {
  				Z_ = logsumexp(Z_, node_[x_.size() - 1][j]->alpha_w_loss, j == 0);
  			}
  			for (size_t i = 0; i < x_.size(); ++i) {
  				for (size_t j = 0; j < ysize_; ++j) {
  					node_[i][j]->prob = std::exp(node_[i][j]->alpha_w_loss + node_[i][j]->beta_w_loss - Z_);
  					//cout<<"node: "<<i<<" "<<j<<" "<<node_[i][j]->prob<<endl;
  					for (const_Path_iterator it = node_[i][j]->lpath.begin(); it != node_[i][j]->lpath.end(); ++it) {
  						(*it)->prob = std::exp((*it)->lnode->alpha_w_loss + (*it)->cost + (*it)->rnode->cost_w_loss + (*it)->rnode->beta_w_loss - Z_);
  						//cout<<"edge: "<<(*it)->lnode->y<<" "<<(*it)->rnode->y<<" "<<(*it)->prob<<endl;
  					}
  				}
  			}
  		} else {
			for (size_t i = 0; i < x_.size(); ++i) {
				double sum_x = 0.0;
				for (size_t j = 0; j < ysize_; ++j) {
					sum_x = logsumexp(sum_x, node_[i][j]->alpha_w_loss, (j == 0));
				}
				Z_ += sum_x;
				for (size_t j = 0; j < ysize_; ++j) {
					node_[i][j]->prob = std::exp(node_[i][j]->alpha_w_loss - sum_x);
				}
			}
		}

		for (size_t i = 0;   i < x_.size(); ++i) {
		  for (size_t j = 0; j < ysize_; ++j) {
			node_[i][j]->calcExpectation(expected, Z_, ysize_);
			double *fv = node_[i][j]->embedding_word_vec;
			for (int v = 0; v < feature_index_->word_vec_size; ++v) {
				double value = *fv;
				expected[v * feature_index_->embedding_label_size + node_[i][j]->y] += value * node_[i][j]->prob;
				++fv;
			}
		  }
		}

#ifdef _DEBUG
       PrintMarginals();
#endif

       /*cout<<"before:"<<endl;
        	  for (int i = 120; i < 124; ++i) {
        			cout<<expected[i]<<" ";
        	  }
        	  cout<<endl;
       */
       double s = 0.0;
	   for (size_t i = 0;   i < x_.size(); ++i) {
		  //cout<<"node "<<i<<" "<<answer_[i]<<":"<<node_[i][answer_[i]]->cost<<" "<<node_[i][answer_[i]]->prob<<endl;
		 for (const int *f = node_[i][answer_[i]]->fvector; *f != -1; ++f) {
			 --expected[*f + answer_[i]];
			 //cout<<*f +answer_[i]<<" "<<expected[*f + answer_[i]]<<endl;
		 }
		double *fv = node_[i][answer_[i]]->embedding_word_vec;
		for (int k = 0; k < feature_index_->word_vec_size; ++k) {
			double value = *fv;
			expected[k*feature_index_->embedding_label_size + answer_[i]] -= value;
			++fv;
		}

		 s += node_[i][answer_[i]]->cost;  // UNIGRAM cost
		 //cout<<answer_[i]<<":"<<node_[i][answer_[i]]->cost<<" ";
		 const std::vector<Path *> &lpath = node_[i][answer_[i]]->lpath;
		 for (const_Path_iterator it = lpath.begin(); it != lpath.end(); ++it) {
		   if ((*it)->lnode->y == answer_[(*it)->lnode->x]) {
			 //cout<<"edge "<<i<<" "<<(*it)->lnode->y<<","<<answer_[i]<<":"<<(*it)->cost<<endl;
			 for (const int *f = (*it)->fvector; *f != -1; ++f) {
			   --expected[*f +(*it)->lnode->y * ysize_ +(*it)->rnode->y];
			   //cout<<*f +(*it)->lnode->y * ysize_ +(*it)->rnode->y<<" "<<expected[*f +(*it)->lnode->y * ysize_ +(*it)->rnode->y]<<endl;
			 }
			 s += (*it)->cost;  // BIGRAM COST
			 break;
		   }
		 }
	   }
	   //cout<<endl;

	  /*cout<<"after:"<<endl;
	  for (int i = 120; i < 124; ++i) {
			cout<<expected[i]<<" ";
	  }
	  cout<<endl;
      */

        //viterbi_decoder.inference(x_.size(), ysize_, node_, result_);  // call for eval()
#ifdef DEBUG_
	   cout<<"Z_ "<<Z_<<" "<<s<<endl;
#endif
        return Z_ - s ;
  }

  void TaggerImpl::forwardbackwardwithConstraints()
  {
	  if (x_.empty()) return;

	  for (int i = 0; i < static_cast<int>(x_.size()); ++i)
	         for (size_t j = 0; j < ysize_; ++j)
	            node_[i][j]->calcAlphaWithConstraints();

	     for (int i = static_cast<int>(x_.size() - 1); i >= 0;  --i)
	 	   for (size_t j = 0; j < ysize_; ++j)
	 		   node_[i][j]->calcBetaWithConstraints();

	    Z_ = 0.0;
	    for (size_t j = 0; j < ysize_; ++j)
	 	 Z_ = logsumexp(Z_, node_[x_.size() - 1][j]->alpha, j == 0);
  }

  void TaggerImpl::computerZ()
  {
	 if (x_.empty()) return;

	 for (int i = 0; i < static_cast<int>(x_.size()); ++i) {
		 for (size_t j = 0; j < ysize_; ++j) {
			 node_[i][j]->calcAlpha();
		 }
	 }

	 for (int i = static_cast<int>(x_.size() - 1); i >= 0;  --i) {
		 for (size_t j = 0; j < ysize_; ++j) {
			 node_[i][j]->calcBeta();
		 }
	 }

	 // calculate node->prob
	 Z_ = 0.0;
	 if (!zero_order) {
		for (size_t j = 0; j < ysize_; ++j) {
			Z_ = logsumexp(Z_, node_[x_.size() - 1][j]->alpha, j == 0);
		}
		for (size_t i = 0; i < x_.size(); ++i) {
			for (size_t j = 0; j < ysize_; ++j) {
				node_[i][j]->prob = std::exp(node_[i][j]->alpha + node_[i][j]->beta - Z_);
				//cout<<"node: "<<i<<" "<<j<<" "<<node_[i][j]->prob<<endl;
				for (const_Path_iterator it = node_[i][j]->lpath.begin(); it != node_[i][j]->lpath.end(); ++it) {
					(*it)->prob = std::exp((*it)->lnode->alpha + (*it)->cost + (*it)->rnode->cost + (*it)->rnode->beta - Z_);
					//cout<<"edge: "<<(*it)->lnode->y<<" "<<(*it)->rnode->y<<" "<<(*it)->prob<<endl;
				}
			}
		}
	 } else {
		for (size_t i = 0; i < x_.size(); ++i) {
			double sum_x = 0.0;
			for (size_t j = 0; j < ysize_; ++j) {
				sum_x = logsumexp(sum_x, node_[i][j]->alpha, (j == 0));
			}
			Z_ += sum_x;
			for (size_t j = 0; j < ysize_; ++j) {
				node_[i][j]->prob = std::exp(node_[i][j]->alpha - sum_x);
			}
		}
	}

    if (!zero_order) {
	  viterbi_decoder.inference(x_.size(), ysize_, node_, result_);
    } else {
	  for (int i = 0; i < x_.size(); ++i) {
		   double maxv = LOG0;
		   for (int y = 0; y < ysize_; ++y) {
				   if (node_[i][y]->cost > maxv) {
					   maxv = node_[i][y]->cost;
					   result_[i] = y;
				   }
		   }
	   }
    }

#ifdef DEBUG_
	 cout<<"alpha"<<endl;
	 for (int i = 0; i < static_cast<int>(x_.size()); ++i) {
	   for (size_t j = 0; j < ysize_; ++j) {
		 cout.precision(20);
		 cout<<node_[i][j]->alpha<<endl;
	   }
	 }
	 //cout<<"beta"<<endl;
	 for (int i = static_cast<int>(x_.size() - 1); i >= 0;  --i) {
	   for (size_t j = 0; j < ysize_; ++j) {
		 cout.precision(20);
		 cout<<node_[i][j]->beta<<" ";
	   }
	   cout<<endl;
	 }
	 cout<<endl;
#endif
  }
  void TaggerImpl::calcCost(Node *n) {
	  feature_index_->calcCost(n);
	  double *fv = n->embedding_word_vec;
	  for (int i = 0; i < feature_index_->word_vec_size; ++i) {
		   double v = *fv;
		   n->cost += feature_index_->get_alpha(i * feature_index_->embedding_label_size + n->y) * v;
		   fv++;
	  }
  }
  void TaggerImpl::recomputeCost()
  {
	  //clear node structure
	  	 for (size_t i = 0; i < x_.size(); ++i) {
	  		  for (size_t j = 0; j < ysize_; ++j)
	  		  {
	  			 node_[i][j]->cost = 0.0;

	  			 vector<Path *> &lpath = node_[i][j]->lpath;
	  			 for (const_Path_iterator it = lpath.begin(); it != lpath.end(); ++it)
	  				 (*it)->cost = 0.0;
	  		  }
	  		}

	  	 for (size_t i = 0; i < x_.size(); ++i) {
	  	   for (size_t j = 0; j < ysize_; ++j)
	  	   {
	  		 calcCost(node_[i][j]);

	  		 const std::vector<Path *> &lpath = node_[i][j]->lpath;
	  		 for (const_Path_iterator it = lpath.begin(); it != lpath.end(); ++it)
	  		   feature_index_->calcCost(*it);
	  	   }
	  	 }
  }

  void TaggerImpl::recomputeCostwithLoss()
   {
	  recomputeCost();

	  double alpha = 1.0;
	  double beta = 1.0;
	  for (size_t i = 0; i < x_.size(); ++i) {
	   for (size_t j = 0; j < ysize_; ++j) {
		   node_[i][j]->cost_w_loss = node_[i][j]->cost;

		   // multiclass precision, recall
		   /*double rec_error = 0.0;
		   double pre_error = 0.0;

		   double error = 0.0;
		   if (answer_[i] != j) {
			   if (answer_[i] != 0 && j != 0) {
				   error += 1.0;
			   } else if (answer_[i] != 0 && j == 0) { //rec
				   error += 10.0;
			   } else if (answer_[i] == 0 && j != 0) { //prec
				   //error += 1.0;
			   }
		   }

		   //node_[i][j]->cost_w_loss += error;
		   //node_[i][j]->cost_w_loss += alpha*pre_error + beta*rec_error;
		    */
	   }
	 }
   }

  void TaggerImpl::label_evaluate() {
       if (x_.empty()) return;

       feature_index_->buildCRFTagger(this);

       recomputeCost();
       computerZ();
  }

  bool TaggerImpl::parse() {
	buildTestFeatureIndex();

	if (x_.empty()) return false;

	Inference();

	return true;
  }
  
  // After computing cost
  void TaggerImpl::Decoding() {
	  computerZ();
  }
  void TaggerImpl::InferencewithNonDuals(vector<double> &u, double sign) {
	  feature_index_->buildCRFTagger(this);
  	  recomputeCost();
  	  int non_label = feature_index_->yID("O");
  	  for (int i = 0; i < x_.size(); ++i) {
  		   node_[i][non_label]->cost += u[i] * sign;
  	  }
  	  Decoding();
  }

  void TaggerImpl::InferencewithDuals(vector<vector<double> > &u, double sign) {
	  feature_index_->buildCRFTagger(this);
	  recomputeCost();
	  for (int i = 0; i < x_.size(); ++i) {
		  for (int y = 0; y < ysize_; ++y) {
			  node_[i][y]->cost += u[i][y] * sign;
		  }
	  }
	  Decoding();
  }

  bool TaggerImpl::Inference() {
	if (x_.empty()) {
		cout<<"Empty sentence!!!"<<endl;
		return false;
	}
	feature_index_->buildCRFTagger(this);
	recomputeCost();
	Decoding();
	return true;
  }

  double TaggerImpl::span_cost(int start, int end, size_t y) {
		double c = node_[start][y]->cost;
		for (int i = start+1; i <= end; ++i) {
			c += node_[i][y]->cost;
			for (const_Path_iterator it = node_[i][y]->lpath.begin();
					it != node_[i][y]->lpath.end(); ++it) {
				if ((*it)->lnode->y == y) {
					c += (*it)->cost;
					break;
				}
			}
		}
		return c;
 }

double TaggerImpl::segment_prob(int start, int end) {
	int segy = result_[start];
	double c = span_cost(start, end, segy);
	double lc = 0.0;
	for (const_Path_iterator it = node_[start][segy]->lpath.begin();
			it != node_[start][segy]->lpath.end(); ++it) {
		lc = logsumexp(lc, (*it)->cost +(*it)->lnode->alpha,
				  (it == node_[start][segy]->lpath.begin()));
	}
	double rc = 0.0;
	for (const_Path_iterator it = node_[end][segy]->rpath.begin();
			it != node_[end][segy]->rpath.end(); ++it) {
		rc = logsumexp(rc, (*it)->rnode->beta + (*it)->rnode->cost + (*it)->cost,
				(it == node_[end][segy]->rpath.begin()));
	}
	return std::exp(lc + c + rc - Z_);
}

void TaggerImpl::span_prob(int start, int end, vector<double> &probs) {
/*	probs.resize(ysize_, 0.0);
	for (int y = 0; y < ysize_; ++y) {
		probs[y] = std::exp(span_cost(start, end, y));
	}
	// renormalize
	double sum = 0.0;
	for (int y = 0; y < ysize_; ++y) {
		sum += probs[y];
	}
	for (int y = 0; y < ysize_; ++y) {
		probs[y] = probs[y]/sum;
	}*/
	probs.resize(ysize_, 0.0);
	int nony = feature_index_->yID("O");

	int opy = feature_index_->yID("OPINION");
	double c = span_cost(start, end, opy);
	double lc = 0.0;
	for (const_Path_iterator it = node_[start][opy]->lpath.begin();
			it != node_[start][opy]->lpath.end(); ++it) {
		lc = logsumexp(lc, (*it)->cost +(*it)->lnode->alpha,
		          (it == node_[start][opy]->lpath.begin()));
	}
	double rc = 0.0;
	for (const_Path_iterator it = node_[end][opy]->rpath.begin();
			it != node_[end][opy]->rpath.end(); ++it) {
		rc = logsumexp(rc, (*it)->rnode->beta + (*it)->rnode->cost + (*it)->cost,
				(it == node_[end][opy]->rpath.begin()));
	}
	probs[opy] = std::exp(lc + c + rc - Z_);
	//double compare = std::exp(node_[start][nony]->alpha+node_[start][nony]->beta-Z_);

	c = span_cost(start, end, nony);
	lc = 0.0;
	for (const_Path_iterator it = node_[start][nony]->lpath.begin();
			it != node_[start][nony]->lpath.end(); ++it) {
		lc = logsumexp(lc, (*it)->cost +(*it)->lnode->alpha,
				  (it == node_[start][nony]->lpath.begin()));
	}
	rc = 0.0;
	for (const_Path_iterator it = node_[end][nony]->rpath.begin();
			it != node_[end][nony]->rpath.end(); ++it) {
		rc = logsumexp(rc, (*it)->rnode->beta + (*it)->rnode->cost + (*it)->cost,
				(it == node_[end][nony]->rpath.begin()));
	}
	probs[nony] = std::exp(lc + c + rc - Z_);

	// renormalized
	double sum = 0.0;
	for (int i = 0; i < probs.size(); ++i) {
		sum += probs[i];
	}
	for (int i = 0; i < probs.size(); ++i) {
		probs[i] = probs[i]/sum;
	}
}

double TaggerImpl::next_transition_cost(size_t i, size_t j, size_t k) {
	for (const_Path_iterator it = node_[i][j]->rpath.begin();
				it != node_[i][j]->rpath.end(); ++it) {
		if ((*it)->rnode->y == k) {
			return (*it)->cost;
		}
	}
	cout<<"Couldn't compute next transition cost!!!"<<endl;
	return -1;
}

double TaggerImpl::prev_transition_cost(size_t i, size_t j, size_t k) {
	for (const_Path_iterator it = node_[i][j]->lpath.begin();
				it != node_[i][j]->lpath.end(); ++it) {
		if ((*it)->lnode->y == k) {
			return (*it)->cost;
		}
	}
	cout<<"Couldn't compute prev transition cost!!!"<<endl;
	return -1;
}

double TaggerImpl::SpanLoss(int start, int end, int y, int y_hat) {
	double c = 0;
	// segment loss
	if (y == y_hat) c = 0;
	else c = 1;
	return c;
}

void TaggerImpl::SegmentInferenceWithLoss() {
  	feature_index_->buildCRFTagger(this);
  	recomputeCost();
  	computerZ();

  	int max_len = 1;
  	int xsize_ = (int)x_.size();
  	vector<vector<vector<SegmentNode*> > >seg_nodes;

  	for (int i = 0; i < xsize_; ++i) {
  		vector<vector<SegmentNode*> > nodes;
		for(int d=0; d<max_len; d++) {
			if (i-d < 0) continue;
			vector<SegmentNode*> cur_nodes;
			vector<double> probs;
			span_prob(i-d, i, probs);
			for (int y = 0; y < ysize_; ++y) {
				SegmentNode *node = new SegmentNode();
				node->start = i-d;
				node->end = i;
				node->y = y;
				node->cost = span_cost(i-d, i, y);
				node->exploss = 0.0;
				node->prob = probs[y];
				cur_nodes.push_back(node);
			}
			nodes.push_back(cur_nodes);
		}
		seg_nodes.push_back(nodes);
  	}

  	// compute exploss and update cost for each node
  	for (int i = 0; i < xsize_; ++i) {
		for(int d=0; d<seg_nodes[i].size(); d++) {
			for (int y = 0; y < ysize_; ++y) {
				SegmentNode *node = seg_nodes[i][d][y];
				double exploss = 0.0;
				for (int y_hat = 0; y_hat < ysize_; ++y_hat) {
					exploss += SpanLoss(node->start, node->end, y, y_hat) * seg_nodes[i][d][y_hat]->prob;
				}
				node->cost = -exploss;
			}
		}
	}

  	// each token select the one with the maximum negative loss
  	for (int i = 0; i < xsize_; ++i) {
  		double maxv = -1e37;
  		int besty = -1;
  		for (int y =0; y < ysize_; ++y) {
  			if (seg_nodes[i][0][y]->cost > maxv) {
  				maxv = seg_nodes[i][0][y]->cost;
  				besty = y;
  			}
  		}
  		result_[i] = besty;
  	}

/*	for (int i = 0; i < xsize_; ++i) {
		for(int d=0; d<seg_nodes[i].size(); d++) {
			for (size_t j = 0; j < ysize_; ++j) {
				double bestc = -1e37;
				SegmentNode *best = 0;
				SegmentNode *node = seg_nodes[i][d][j];
				if (node->start-1 < 0) continue;
				for (int pd = 0; pd < seg_nodes[node->start-1].size(); ++pd) {
				   for (int prey = 0; prey < ysize_; ++prey) {
					  //double cost = seg_nodes[node->start-1][pd][prey]->bestCost
					  //	  + next_transition_cost(node->start-1, prey, j) + node->cost;
					  double cost = seg_nodes[node->start-1][pd][prey]->bestCost + node->cost;
					  if (cost > bestc) {
						  bestc = cost;
						  best  = seg_nodes[node->start-1][pd][prey];
					  }
					}
				}
				node->prev     = best;
				node->bestCost = best ? bestc : node->cost;
			}
		}
	}

	double bestc = -1e37;
	SegmentNode *best = 0;
	size_t s = xsize_-1;
	for (int ns = 0; ns < seg_nodes[s].size(); ++ns) {
		for (int j = 0; j < ysize_; ++j) {
			 if (bestc < seg_nodes[s][ns][j]->bestCost) {
			   best  = seg_nodes[s][ns][j];
			   bestc = seg_nodes[s][ns][j]->bestCost;
			 }
		}
	}

	for (SegmentNode *n = best; n; n = n->prev) {
		for (int i = n->start; i <= n->end; ++i) {
			 result_[i] = n->y;
		}
	}

	viterbi_decoder.cost_ = -best->bestCost;
	*/
 }

  bool TaggerImpl::parse_stream(std::istream *is, std::ostream *os) {
    if (!read(is) || !parse()) return false;
    if (x_.empty()) return true;
    toString();
    os->write(os_.data(), os_.size());
    return true;
  }

  bool TaggerImpl::buildTestFeatureIndex() {
  	if (x_.empty()) return true;

  	feature_index_->buildCRFFeatures(this, false);
  	feature_index_->buildCRFTagger(this);

    return true;
  }

  bool TaggerImpl::next() {
	  return viterbi_decoder.next(result_);
  }

  string TaggerImpl::toString()
  {
    os_.assign("");

	if(viterbi_decoder.nbest_ > 1) {
		vector<string> results(x_.size());
		fill(results.begin(), results.end(), "");

		vector<double> probs;
		int n = 0;
		while(n < viterbi_decoder.nbest_) {
			if (!viterbi_decoder.next(result_)) break;
			
			double p = prob();
			probs.push_back(p);

			for (size_t i = 0; i < x_.size(); ++i) {
				string str = yname(y(i));
				results[i] += str + "\t";
			}
			n++;
		}

		for (size_t i = 0; i < x_.size(); ++i) {
			string label = yname(answer_[i]);
		    os_ << x_[i][0] <<"\t"<<label<<"\t"<<results[i]<<"\n";
		}      

		/*for (size_t n = 0; n < probs.size(); ++n)  {
			//double p = (double)probs[n]/sum;
			double p = (double)probs[n];
			os_<<p<<"\t";
		}
		os_<<"\n";*/
		os_ << "\n";
	} else {
		double p = prob();
		for (size_t i = 0; i < x_.size(); ++i) {
			os_ << x_[i][0] <<"\t"<<yname(answer_[i])<<"\t";
			os_ << yname(y(i));
            os_ << '\n';    
		}
		os_ << '\n';
    }                                                             
	
	return os_.c_str();
  }

  bool TaggerImpl::outputPrediction(std::ostream *os, vector<string>& results)
   {
 	  if (x_.empty()) return true;

 	  for (size_t i = 0; i < x_.size(); ++i)
 		  results.push_back(yname(y(i)));

 	  toString();
 	  os->write(os_.data(), os_.size());
 	  return true;
   }

  bool TaggerImpl::resulttoString( vector<string>& results)
     {
   	  if (x_.empty()) return true;

   	  for (size_t i = 0; i < x_.size(); ++i)
   		  results.push_back(yname(y(i)));

   	  return true;
     }

    double TaggerImpl::stateEnergy(int pos, unsigned int k, vector<unsigned short int> y) {
    	    return 1.0;
    }
	double TaggerImpl::ComputeEnergyChange(vector<unsigned short int> y, int var_index,
		  unsigned short int old_state, unsigned short int new_state) {
		return 1.0;
	}
	double TaggerImpl::EvaluateEnergy(vector<unsigned short int> y) {
		return 1.0;
	}
	void TaggerImpl::getconditional(int pos, vector<unsigned short int> y, vector<double>& p, double temp) {

	}

	void TaggerImpl::PrintMarginals() {
		for (size_t i = 0;   i < x_.size(); ++i) {
			/*cout<<x_[i][0]<<"\t";
			for (int y = 0; y < ysize_; ++y) {
				cout<<y<<":"<<node_[i][y]->prob<<" ";
			}
			cout<<endl;*/

			for (int y = 0; y < ysize_; ++y) {
			  const std::vector<Path *> &lpath = node_[i][y]->lpath;
			  if (lpath.size() > 0) {
				  cout<<x_[lpath[0]->lnode->x][0]<<"->"<<x_[i][0]<<"\t";
			  }
			  for (const_Path_iterator it = lpath.begin(); it != lpath.end(); ++it) {
				cout<<(*it)->lnode->y<<","<<y<<":";
				cout<<(*it)->prob<<" ";
			  }
			}
			cout<<endl;
		}
	}

	void TaggerImpl::PrintParameters() {
		for (size_t i = 0;   i < x_.size(); ++i) {
			cout<<x_[i][0]<<"\t";
			for (int y = 0; y < ysize_; ++y) {
				for (int *f = node_[i][y]->fvector; *f != -1; ++f) {
					cout<<feature_index_->get_alpha(*f + y)<<" ";
				}
			}
			cout<<endl;

			for (int y = 0; y < ysize_; ++y) {
			  const std::vector<Path *> &lpath = node_[i][y]->lpath;
			  if (lpath.size() > 0) {
				  cout<<x_[lpath[0]->lnode->x][0]<<"->"<<x_[i][0]<<"\t";
			  }
			  for (const_Path_iterator it = lpath.begin(); it != lpath.end(); ++it) {
				  for (int *f = (*it)->fvector; *f != -1; ++f) {
					  int fid = *f + (*it)->lnode->y * ysize_ + y;
					  cout<<feature_index_->get_alpha(fid)<<" ";
				  }
			  }
			}
			cout<<endl;
		}
	}

bool TaggerImpl::SentenceToTagger(Sentence *s, CRFPP::FeatureIndex *feature_index, bool train) {
	sent = s;

	open(feature_index);

	int length = sent->size();
	if (length == 0) return false;

	clear();

	answer_.resize(length, 0);
	result_.resize(length, 0);

	// read labels
	for(int i=0; i<sent->token_labels.size(); i++) {
		string label = sent->token_labels[i];
		if (feature_index->lb2i.find(label) == feature_index->lb2i.end()) {
			 cout<<"cannot find answer: " << label <<endl;
			 return false;
		}
		answer_[i] = feature_index->lb2i[label];
	}

	x_.resize(length);
	for (int i = 0; i < sent->words.size(); ++i) {
		x_[i].push_back(sent->words[i].c_str());
	}
	node_.resize(length);
	for (int i = 0; i < length; ++i) {
		node_[i].resize(ysize_);
	}

	feature_index_->buildCRFFeatures(this, train);

	return true;
}

}


