/*
 * CRFLearner.cpp
 *
 *  Created on: Jan 18, 2013
 *      Author: bishan
 */

#include "CRFLearner.h"
#include "./Parsing/Utils.h"
#include "./CRF/feature_index.h"
#include "TaggingEval.h"
#include <iostream>

CRFLearner::~CRFLearner() {
	// TODO Auto-generated destructor stub
}

void CRFLearner::BuildSentenceTagger() {
	encoder.labelx.clear();
	for (int i = 0; i < trainset->sentences.size(); ++i) {
		CRFPP::TaggerImpl* x  = new CRFPP::TaggerImpl();
		if (x->SentenceToTagger(trainset->sentences[i], encoder.feature_index, true)) {
			int xid = encoder.labelx.size();
			x->set_thread_id(xid % encoder.thread_num);
			encoder.labelx.push_back(x);
		} else {
			delete x;
			x = NULL;
		}
	}

	cout<<"Loaded "<<encoder.labelx.size()<<" instances..."<<endl;
}

void CRFLearner::InitializeEncoder(CRFPP::FeatureIndex *p_feature_index) {
	encoder.feature_index = (CRFPP::EncoderFeatureIndex*)p_feature_index;
}

void CRFLearner::TrainEncoder(string modelfile) {
	cout<<"Feature size before shrinking "<<encoder.feature_index->size()<<endl;
	encoder.feature_index->shrink(encoder.freq);
	cout<<"Feature size after shrinking "<<encoder.feature_index->size()<<endl;

	//encoder.feature_index->BuildFeatureVec();

	// allocate nodes, paths, and fvectors
	encoder.RebuildAlpha();

	if(sgd) {
		encoder.SGD_learn(modelfile, learningrate, decayRate, sgd_iteration);
	} else {
	    encoder.learn(modelfile.c_str());
	}
}

void CRFPredictor::LoadModel(string modelfile) {
	feature_index = new CRFPP::DecoderFeatureIndex();
	feature_index->open(modelfile.c_str());
}

void CRFPredictor::BuildSentenceTagger() {
	testx.clear();
	test_sentences.clear();
	for (int i = 0; i < testset->sentences.size(); ++i) {
		CRFPP::TaggerImpl* x = new CRFPP::TaggerImpl();
		if (x->SentenceToTagger(testset->sentences[i], feature_index, false)) {
			x->viterbi_decoder.nbest_ = nbest;
			testx.push_back(x);
			test_sentences.push_back(testset->sentences[i]);
		} else {
			delete x;
			x = NULL;
		}
	}
	cout<<"Loaded "<<testx.size()<<" instances"<<endl;
}

void CRFPredictor::NBestPrediction(string resultfile) {
	ifstream infile(resultfile.c_str(), ios::in);
	vector<string> results;
	string sent_result = "";
	string str;
	while (getline(infile, str)) {
		if (str == "") {
			results.push_back(sent_result);
			sent_result = "";
		} else {
			sent_result += str + "\n";
		}
	}
	infile.close();

	int n = 0;
	double alltruep = 0;
	double allpredictp = 0;
	double allpropp = 0;
	double allpropr = 0;

	for (int i = 0; i < results.size(); ++i) {
		vector<string> lines;
		Utils::Split(results[i], '\n', lines);
		vector<vector<string> > values;
		for (int k = 0; k < lines.size(); ++k) {
			vector<string> fields;
			Utils::Split(lines[k], '\t', fields);
			values.push_back(fields);
		}

		int ncol = values[0].size();
		if (ncol < 3) continue;

		// read gold standard labels
		vector<Span> answers;
		int index = 1;
		int p = 0;
		while (p < values.size()) {
			if (values[p][index] != "O") {
				int start = p;
				p++;
				while (p < values.size() && values[p][index] == values[start][index]) {
					p++;
				}
				int end = p-1;
				Span eval_en;
				eval_en.start = start;
				eval_en.end = end;
				eval_en.label = values[start][index];
				answers.push_back(eval_en);
			} else {
				p++;
			}
		}

		vector<vector<Span> > nbest_results(nbest);
		for (int k = 0; k < nbest; ++k) {
			if (k+2 >= ncol) continue;
			int index = k+2;
			int p = 0;
			while (p < values.size()) {
				if (values[p][index] != "O") {
					int start = p;
					p++;
					while (p < values.size() && values[p][index] == values[start][index]) {
						p++;
					}
					int end = p-1;
					Span eval_en;
					eval_en.start = start;
					eval_en.end = end;
					eval_en.label = values[start][index];
					nbest_results[k].push_back(eval_en);
				} else {
					p++;
				}
			}
		}

		// pick the best overlapping prediction
		double truep = answers.size();
		double best_index = 0;
		double best_predictp = 0.0;
		double best_propr = 0.0;
		double best_propp = 0.0;
		double best_f1 = 0.0;
		for (int k = 0; k < nbest_results.size(); ++k) {
			double propr = 0.0;
			double propp = 0.0;
			if (answers.size() == 0 && nbest_results[k].size() == 0) continue;

			double predictp = nbest_results[k].size();
			ProportionMatch(answers, nbest_results[k], propr, propp);
			double pre = propp==0 ? 0 : propp/predictp;
			double rec = propr==0 ? 0 : propr/truep;
			double f = pre+rec==0 ? 0 : 2.0*pre*rec/(pre+rec);
			if (f > best_f1) {
				best_f1 = f;
				best_predictp = predictp;
				best_propr = propr;
				best_propp = propp;

				best_index = k;
			}
		}

		alltruep += truep;
		allpredictp += best_predictp;
		allpropp += best_propp;
		allpropr += best_propr;

		n++;
	}

	double pre = allpropp==0 ? 0 : allpropp/allpredictp;
	double rec = allpropr==0 ? 0 : allpropr/alltruep;
	double f = pre+rec==0 ? 0 : 2.0*pre*rec/(pre+rec);

	cout<<"pre: "<<pre<<" rec: "<<rec<<" f: "<<f<<endl;
}

void CRFPredictor::TaggingPrediction(string filename) {
	// Evaluation
	cout<<"test instance size: "<<testx.size()<<endl;
	ofstream errorfile(filename.c_str(), ios::out);

	// read normal results
	vector<vector<string> > answers;
	vector<vector<string> > results;

	for(int i=0; i<testx.size(); ++i) {
		CRFPP::TaggerImpl *tagger = testx[i];
		testx[i]->Inference();

		vector<string> labels;
		vector<string> pred_labels;
		for(int j=0; j<tagger->size(); j++) {
			int itrue = tagger->answer_[j];
			int ipred = tagger->result_[j];

			string pred_label = feature_index->ystr(ipred);
			string label = feature_index->ystr(itrue);

			labels.push_back(label);
			pred_labels.push_back(pred_label);
		}

		answers.push_back(labels);
		results.push_back(pred_labels);

		// set prediction results in sentence
		errorfile<<tagger->toString();
	}

	std::stringstream ss;
	ComputeScore(answers, results, eval_param, ss);

	errorfile<<ss.str()<<endl;
	errorfile.close();
}
