/*
 * CRFLearner.h
 *
 *  Created on: Jan 18, 2013
 *      Author: bishan
 */

#ifndef CRFLEARNER_H_
#define CRFLEARNER_H_
#include "./CRF/encoder.h"
#include "./CRF/tagger.h"
#include "Dataset.h"
#include "Param.h"

#include "TaggingEval.h"

class CRFLearner {
public:
	CRFLearner() : trainset(NULL) {
		sgd = false;
		adagrad = false;
		learningrate = 0.1;
		decayRate = 0.01;
		zeroorder = false;
		windowsize = 1;
	}
	~CRFLearner();
public:
	Dataset *trainset;

	bool sgd;
	int sgd_iteration;
	double learningrate;
	double decayRate;
	bool adagrad;
	bool zeroorder;
	int  windowsize;

	CRFPP::Encoder encoder;

public:
    void initParam(Param param) {
		sgd = param.use_sgd_learn;
		sgd_iteration = param.sgd_iteration;
		learningrate = param.learningrate;

		encoder.freq = param.freq;
		encoder.thread_num = param.threadnum;
		zeroorder = param.zeroorder;
		windowsize = param.windowsize;
		encoder.C = param.regularizer;
	}
    void InitializeEncoder(CRFPP::FeatureIndex *p_feature_index);
	void TrainEncoder(string modelfile);
	void BuildSentenceTagger();
};

class CRFPredictor {
public:
	CRFPredictor() {
		testset = new Dataset();
		feature_index = NULL;
		nbest = 1;
	}
	~CRFPredictor() {
		for (int i = 0; i < testx.size(); ++i) {
			delete testx[i];
			testx[i] = NULL;
		}
	}
public:
	Dataset *testset;
	CRFPP::DecoderFeatureIndex *feature_index;
	vector<CRFPP::TaggerImpl *> testx;
	vector<Sentence*> test_sentences;
	int nbest;
	EvalParam eval_param;
public:
	void InitializeDecoder(CRFPP::FeatureIndex *p_feature_index) {
		feature_index = (CRFPP::DecoderFeatureIndex*)p_feature_index;
	}
	void LoadModel(string modelfile);
	void TaggingPrediction(string outputfile);

	void BuildSentenceTagger();
	void NBestPrediction(string resultfile);
};


#endif /* CRFLEARNER_H_ */
