/*
 * Dictionary.cpp
 *
 *  Created on: Mar 19, 2013
 *      Author: bishan
 */

#include "Dictionary.h"
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>

//struct passwd *dict_pw = getpwuid(getuid());
//string dict_homedir = dict_pw->pw_dir;
string dict_path = "./resources";

Dictionary::Dictionary() {
	// TODO Auto-generated constructor stub
}

Dictionary::~Dictionary() {
	// TODO Auto-generated destructor stub
}

set<string> Dictionary::SentenceTags = Dictionary::InitializeSentTags();
set<string> Dictionary::PhraseTags = Dictionary::InitializePhraseTags();
set<string> Dictionary::WordTags = Dictionary::InitializeWordTags();

set<string> Dictionary::copulativeVerbs = Dictionary::loadDictionaries(dict_path+"/copulative_verbs");
set<string> Dictionary::rev_connectives = Dictionary::loadDictionaries(dict_path+"/rev_connectives");
set<string> Dictionary::connectives = Dictionary::loadDictionaries(dict_path+"/connectives");
set<string> Dictionary::FunctionTags = Dictionary::loadDictionaries(dict_path+"/functiontags");
set<string> Dictionary::clausalVerbs = Dictionary::loadDictionaries(dict_path+"/clausal_verbs");
set<string> Dictionary::negationWords = Dictionary::loadDictionaries(dict_path+"/negations");

set<string> Dictionary::intensifiers = Dictionary::loadDictionaries(dict_path+"/intensifiers");
set<string> Dictionary::strongmodal = Dictionary::loadDictionaries(dict_path+"/strongmodal");
set<string> Dictionary::weakmodal = Dictionary::loadDictionaries(dict_path+"/weakmodal");
set<string> Dictionary::diminisher = Dictionary::loadDictionaries(dict_path+"/diminisher");

set<string> Dictionary::stopWords = Dictionary::loadDictionaries(dict_path+"/stopwords");
set<string> Dictionary::reportVerbs = Dictionary::loadDictionaries(dict_path+"/reportverbs");
map<string, pair<string, string> > Dictionary::polarities = Dictionary::loadSubjLexicon(dict_path+"/subjclueslen1-HLTEMNLP05.tff");
map<string, string> Dictionary::oplexicon = Dictionary::loadOpLexicon(dict_path+"/positive-words.txt", dict_path+"/negative-words.txt");

map<string, string> Dictionary::framenet_features = Dictionary::InitializeVerbFrames(dict_path+"/verbframes.dict");
map<string, string> Dictionary::lemmas = Dictionary::loadMap(dict_path+"/lemma");
map<string, WordInfo *> Dictionary::vocabulary = Dictionary::LoadVocabulary(dict_path+"/vocabulary.info");
map<string, double> Dictionary::wordnet_distance = Dictionary::LoadWordnetDistance(dict_path+"/wordnet.dist");

// All keywords are lower form of word + "/" + pos[0], e.g. doctor-n

map<string, WordInfo *> Dictionary::LoadVocabulary(string filename) {
	map<string, WordInfo *> vocab;
	ifstream infile(filename.c_str(), ios::in);
	string line;
	while(getline(infile, line)) {
		vector<string> fields;
		Utils::Split(line, '\t', fields);
		// ???? check
		if (fields.size() < 6) continue;

		string key_word = fields[0];
		WordInfo *info = new WordInfo();
		info->lemma = fields[1];
		vector<string> synonyms;
		Utils::Split(fields[2], ',', synonyms);
		// ???? check
		if (synonyms[0] != "-") {
			copy(synonyms.begin(), synonyms.end(), inserter(info->synonyms, info->synonyms.end()));
		}
		vector<string> hypernyms;
		Utils::Split(fields[3], ',', hypernyms);
		if (hypernyms[0] != "-") {
			copy(hypernyms.begin(), hypernyms.end(), inserter(info->hypernyms, info->hypernyms.end()));
		}
		vector<string> n_deriveds;
		Utils::Split(fields[4], ',', n_deriveds);
		if (n_deriveds[0] != "-") {
			copy(n_deriveds.begin(), n_deriveds.end(), inserter(info->noun_derivedforms, info->noun_derivedforms.end()));
		}
		vector<string> v_deriveds;
		Utils::Split(fields[5], ',', v_deriveds);
		if (v_deriveds[0] != "-") {
			copy(v_deriveds.begin(), v_deriveds.end(), inserter(info->verb_derivedforms, info->verb_derivedforms.end()));
		}

		vocab[key_word] = info;
	}
	return vocab;
}

map<string, double> Dictionary::LoadWordnetDistance(string filename) {
	map<string, double> dist;
	ifstream infile(filename.c_str(), ios::in);
	string line;
	while(getline(infile, line)) {
		dist[line] = 1;
	}
	infile.close();
	return dist;
}

void Dictionary::updateOpLexicon(vector<string> &positive_words, vector<string> &pos_pos,
		vector<string> &negative_words, vector<string> &neg_pos)
{
	int common = 0;
	set<string> wordmap;
/*	map<string, string>::iterator iter;
	for (iter=polarities.begin(); iter!=polarities.end(); iter++) {
		string key = iter->first;
		int index = key.find('/');
		string word = key.substr(0,index);
		if (oplexicon.find(word) != oplexicon.end())
			common++;
		wordmap.insert(word);
	}
	*/
	cout<<"common words "<<common<<endl;
	cout<<"mpqa "<<wordmap.size()<<" liu "<<oplexicon.size()<<" mpqa "<<polarities.size()<<endl;

/*	oplexicon.clear();
	polarities.clear();
	for (int i = 0; i < positive_words.size(); ++i) {
		string word = Utils::toLower(positive_words[i]);
		string pos = pos_pos[i];
		string lemma = Dictionary::getLemma(word, pos);

		oplexicon[word] = "positive";
		oplexicon[lemma] = "positive";

		polarities[word+"/"+pos[0]] = "positive";
		polarities[lemma+"/"+pos[0]] = "positive";
	}

	for (int i = 0; i < negative_words.size(); ++i) {
		string word = Utils::toLower(negative_words[i]);
		string pos = neg_pos[i];
		string lemma = Dictionary::getLemma(word, pos);

		oplexicon[word] = "negative";
		oplexicon[lemma] = "negative";

		polarities[word+"/"+pos[0]] = "negative";
		polarities[lemma+"/"+pos[0]] = "negative";
	}*/
}

map<string, string> Dictionary::loadOpLexicon(string posfilename, string negfilename)
{
	map<string, string> lexicon;
	ifstream posfile(posfilename.c_str(), ios::in);
	if(!posfile)
		return lexicon;
	string str;
	while(getline(posfile, str))
	{
		if(str == "")
			continue;
		if(str[0] == ';')
			continue;
		lexicon[str] = "positive";
	}
	posfile.close();

	ifstream negfile(negfilename.c_str(), ios::in);
	if(!negfile)
		return lexicon;
	while(getline(negfile, str))
	{
		if(str == "")
			continue;
		if(str[0] == ';')
			continue;
		lexicon[str] = "negative";
	}
	negfile.close();

	return lexicon;
}

map<string, string> Dictionary::loadMap(string filename)
{
	map<string, string> dict;
	ifstream infile(filename.c_str(), ios::in);
	string str;
	while(getline(infile, str)) {
		vector<string> fields;
		Utils::Split(str, '\t', fields);
		string key = Utils::toLower(fields[0]);
		dict[key] = fields[1];
	}
	infile.close();
	return dict;
}

map<string, pair<string, string> > Dictionary::loadSubjLexicon(string filename)
{
	map<string, pair<string, string> > pmap;
	ifstream infile(filename.c_str(), ios::in);
	if(!infile)
		return pmap;
	string str;
	while(getline(infile, str))
	{
		vector<string> fields;
		Utils::Split(str, ' ', fields);

		string word = Utils::getValue(fields, "word1", "=");
		string pos = Utils::getValue(fields, "pos1", "=");
		if(pos == "verb")
			pos = "v";
		else if(pos == "adj")
			pos = "j";
		else if(pos == "noun")
			pos = "n";
		else if(pos == "adverb")
			pos = "r";
		else
			continue; //invalid pos

		string type = Utils::getValue(fields, "type", "=");
		string polarity = Utils::getValue(fields, "priorpolarity", "=");
		//if(polarity == "neutral" || polarity == "positive" || polarity == "negative")
		pmap[word+"/"+pos] = make_pair(type, polarity);
	}
	infile.close();

	return pmap;
}

string Dictionary::getLemma(string word, string pos)
{
	word = Utils::toLower(word);
	string key = word+"/"+pos[0];
	if(lemmas.find(key) != lemmas.end() && lemmas[key] != "")
		return lemmas[key];
	return word;
}

bool Dictionary::HasVerbDerivedForm(string word, string pos) {
	string keyword = Utils::toLower(word) + "/" + pos[0];
	if (vocabulary.find(keyword) != vocabulary.end()) {
		set<string> derivedform = vocabulary[keyword]->verb_derivedforms;
		for (set<string>::iterator iter = derivedform.begin(); iter != derivedform.end(); ++iter) {
			string word = *iter;
			if (word.substr(word.size()-2) == "/V") {
				return true;
			}
		}
	}
	return false;
}

bool Dictionary::IsWordnetSynonym(string word1, string pos1, string word2, string pos2) {
	string keyword_a = Utils::toLower(word1) + "/" + pos1[0];
	string keyword_b = Utils::toLower(word2) + "/" + pos2[0];
	if (keyword_a == keyword_b) return true;
	if (getLemma(word1, pos1) == getLemma(word2, pos2)) return true;
	return (wordnet_distance.find(keyword_a+","+keyword_b) != wordnet_distance.end());
}

string Dictionary::getPolarity(string word, string lemma, string pos)
{
	// every input should be lower case.
	word = Utils::toLower(word);

	if (word == "pros") {
		return "positive";
	} else if (word == "cons") {
		return "negative";
	} else if (word == "pros+cons") {
		return "neutral";
	}

	//use mpqa lexicon
	if (pos.size() > 0) {
		pos = Utils::toLower(pos);
		string key = word+"/"+pos[0];
		if(polarities.find(key) != polarities.end())
			return polarities[key].second;
	} else {
		pos = "j";
		string key = word+"/"+pos[0];
		if(polarities.find(key) != polarities.end())
			return polarities[key].second;

		pos = "r";
		key = word+"/"+pos[0];
		if(polarities.find(key) != polarities.end())
			return polarities[key].second;

		pos = "v";
		key = word+"/"+pos[0];
		if(polarities.find(key) != polarities.end())
			return polarities[key].second;

		pos = "n";
		key = word+"/"+pos[0];
		if(polarities.find(key) != polarities.end())
			return polarities[key].second;
	}

	if (lemma != "") {
		string key = lemma+"/"+pos[0];
		if(polarities.find(key) != polarities.end())
			return polarities[key].second;
	}

	//using product review lexicon
	//use opinion lexicon
/*	if(oplexicon.find(word) != oplexicon.end())
		return oplexicon[word];
	if(oplexicon.find(lemma) != oplexicon.end())
		return oplexicon[lemma];
*/
	return "";

}

string Dictionary::getPolarStrength(string word, string lemma, string pos)
{
	// every input should be lower case.
	word = Utils::toLower(word);

	if (word == "pros" || word == "cons" || word == "pros+cons") {
		return "strongsubj";
	}

	if (pos.size() > 0) {
		pos = Utils::toLower(pos);
		//use mpqa lexicon
		string key = word+"/"+pos[0];
		if(polarities.find(key) != polarities.end())
			return polarities[key].first;
	} else {
		pos = "j";
		string key = word+"/"+pos[0];
		if(polarities.find(key) != polarities.end())
			return polarities[key].first;

		pos = "r";
		key = word+"/"+pos[0];
		if(polarities.find(key) != polarities.end())
			return polarities[key].first;

		pos = "v";
		key = word+"/"+pos[0];
		if(polarities.find(key) != polarities.end())
			return polarities[key].first;

		pos = "n";
		key = word+"/"+pos[0];
		if(polarities.find(key) != polarities.end())
			return polarities[key].first;
	}

	if (lemma != "") {
		string key = lemma+"/"+pos[0];
		if(polarities.find(key) != polarities.end())
			return polarities[key].first;
	}

	return "";
}


set<string> Dictionary::loadDictionaries(string filename)
{
	//cout<<"loading dict "<<filename<<endl;
	set<string> verbs;
	ifstream infile(filename.c_str(), ios::in);
	string str;
	while(getline(infile, str))
		verbs.insert(str);
	infile.close();
	return verbs;
}

map<string, string> Dictionary::InitializeVerbFrames(string filename)
{
	map<string, string> vmap;
	string str;
	ifstream infile(filename.c_str(), ios::in);
	while(getline(infile, str))
	{
		vector<string> fields;
		Utils::Split(str, '\t', fields);
		string value = fields[1];
		if(fields.size() > 2) //semantic info
			value += '\t'+fields[2];
		vmap[fields[0]] = value;
	}
	infile.close();
	return vmap;
}

bool Dictionary::isCopulativeVerb(string word) {
	word = Utils::toLower(word);
	return (copulativeVerbs.find(word) != copulativeVerbs.end());
}

bool Dictionary::isConnective(string word)
{
	word = Utils::toLower(word);
	if(connectives.find(word) != connectives.end())
		return true;
	else
		return false;
}

bool Dictionary::isRevConnective(string word)
{
	word = Utils::toLower(word);
	if(rev_connectives.find(word) != rev_connectives.end())
		return true;
	else
		return false;
}

bool Dictionary::isClausalVerb(string verb)
{
	if(clausalVerbs.find(getVerbLemma(verb)) != clausalVerbs.end())
		return true;
	else
		return false;
}

string Dictionary:: getVerbLemma(string verb)
{
	string key = verb+"/V";
	if(framenet_features.find(key) == framenet_features.end())
		return "";
	string frame = framenet_features[key];
	int index = frame.find('\t');
	return frame.substr(0,index);
}

bool Dictionary::isFunctionWord(string pos)
{
	if(FunctionTags.find(pos) != FunctionTags.end())
		return true;
	else
		return false;
}

void Dictionary::getFrameFeatures(string verb, vector<string>& strfeatures)
{
	string key = verb+"/V";
	if(framenet_features.find(key) == framenet_features.end())
		return;
	string frame = framenet_features[key];
	int index = frame.find('\t');
	string fstr = frame.substr(index+1);

	//parse the frame string
	map<string, int> features;
	string str = "";
	for(int i=0; i<fstr.size(); i++)
	{
		if(fstr[i] == '{')
			continue;
		else if(fstr[i] == '}')
		{
			features[str] = 1; //insert one role set string
			str = "";
		}
		else
			str += fstr[i];
	}
	map<string, int>::iterator iter;
	for(iter=features.begin(); iter!=features.end(); iter++)
	{
		string str = iter->first;
		int i = 0;
		while(i<str.size())
		{
			if(str[i] == '#')
			{
				i++;
				continue;
			}
			if(str[i] >= '0' && str[i]<='9') //argument
			{
				int argn = str[i]-'0';
				string fstr = "arg"+Utils::int2string(argn);
				strfeatures.push_back(fstr);
				i++;
				if(i<str.size() && str[i] == ';')
					i++; //skip ';'
				if(i<str.size() && str[i] != '#')
				{
					string type = "";
					while(i<str.size() && str[i] != '#')
					{
						type+=str[i];
						i++;
					}
					if(type != "")
						strfeatures.push_back(type);
				}
			}
			else
				i++;
		}
	}
}

set<string> Dictionary::InitializeWordTags()
{
	set<string> WordTags;

	WordTags.insert("CC");
	WordTags.insert("CD");
	WordTags.insert("DT");
	WordTags.insert("EX");
	WordTags.insert("FW");
	WordTags.insert("IN");
	WordTags.insert("JJ");
	WordTags.insert("JJR");
	WordTags.insert("JJS");
	WordTags.insert("LS");
	WordTags.insert("MD");
	WordTags.insert("NN");
	WordTags.insert("NNS");
	WordTags.insert("NNP");
	WordTags.insert("NNPS");
	WordTags.insert("PDT");
	WordTags.insert("POS");
	WordTags.insert("PRP");
	WordTags.insert("PRP$");
	WordTags.insert("RB");
	WordTags.insert("RBR");
	WordTags.insert("RBS");
	WordTags.insert("RP");
	WordTags.insert("SYM");
	WordTags.insert("TO");
	WordTags.insert("UH");
	WordTags.insert("VB");
	WordTags.insert("VBD");
	WordTags.insert("VBG");
	WordTags.insert("VBN");
	WordTags.insert("VBP");
	WordTags.insert("VBZ");
	WordTags.insert("WDT");
	WordTags.insert("WP");
	WordTags.insert("WP$");
	WordTags.insert("WRB");

	return WordTags;
}
set<string> Dictionary::InitializePhraseTags()
{
	set<string> PhraseTags;

	PhraseTags.insert("ADJP");
	PhraseTags.insert("ADVP");
	PhraseTags.insert("CONJP");
	PhraseTags.insert("PP");
	PhraseTags.insert("VP");
	PhraseTags.insert("NP");
	PhraseTags.insert("WHADJP");
	PhraseTags.insert("WHAVP");
	PhraseTags.insert("WHNP");
	PhraseTags.insert("WHPP");
	PhraseTags.insert("FRAG");
	PhraseTags.insert("INTJ");
	PhraseTags.insert("LST");
	PhraseTags.insert("NAC");
	PhraseTags.insert("NX");
	PhraseTags.insert("PRN");
	PhraseTags.insert("PRT");
	PhraseTags.insert("QP");
	PhraseTags.insert("RRC");
	PhraseTags.insert("UCP");
	PhraseTags.insert("X");

	return PhraseTags;
}

set<string> Dictionary::InitializeSentTags()
{
	set<string> SentenceTags;

	SentenceTags.insert("ROOT");
	SentenceTags.insert("S");
	SentenceTags.insert("SBAR");
	SentenceTags.insert("SBARQ");
	SentenceTags.insert("SINV");
	SentenceTags.insert("SQ");

	return SentenceTags;
}
