/*
 * Dictionary.h
 *
 *  Created on: Mar 19, 2013
 *      Author: bishan
 */

#ifndef DICTIONARY_H_
#define DICTIONARY_H_

#include <set>
#include <map>
#include <string>

#include "Utils.h"

using namespace std;

enum MentionType { PRONOMINAL, NOMINAL, PROPER };
enum Gender { MALE, FEMALE, NEUTRAL, GENDER_UNKNOWN };
enum Number { SINGULAR, PLURAL, NUMBER_UNKNOWN };
enum Animacy { ANIMATE, INANIMATE, ANIMACY_UNKNOWN };
enum Person { I, YOU, HE, SHE, WE, THEY, IT, PERSON_UNKNOWN};

struct WordInfo {
	string lemma;
	set<string> synonyms;
	set<string> hypernyms;
	set<string> noun_derivedforms;
	set<string> verb_derivedforms;
};

class Dictionary {
public:
	Dictionary();
	virtual ~Dictionary();

public:
	static set<string> SentenceTags;
    static set<string> PhraseTags;
    static set<string> WordTags;

    static set<string> clausalVerbs;
    static map<string, string> framenet_features;
    static set<string> negationWords;
    static set<string> reportVerbs;
    static set<string> stopWords;

    static set<string> intensifiers;
    static set<string> strongmodal;
    static set<string> weakmodal;
    static set<string> diminisher;

    static set<string> FunctionTags;

    static map<string, string> oplexicon;
    static map<string, pair<string, string> > polarities;
    static map<string, string> lemmas;

    static set<string> rev_connectives;
    static set<string> connectives;

    static set<string> copulativeVerbs;

    static map<string, double> wordnet_distance;
    	static map<string, WordInfo *> vocabulary;

public:
    	static map<string, WordInfo *> LoadVocabulary(string filename);
    static map<string, double> LoadWordnetDistance(string filename);

    static bool isRevConnective(string word);
    static bool isConnective(string word);
    static bool isCopulativeVerb(string word);

    static map<string, string> loadMap(string filename);
    static map<string, pair<string, string> > loadSubjLexicon(string filename);
    static map<string, string> loadOpLexicon(string posfilename, string negfilename);

    static set<string> loadDictionaries(string filename);

	static set<string> InitializeSentTags();
	static set<string> InitializeWordTags();
	static set<string> InitializePhraseTags();

	static map<string, string> InitializeVerbFrames(string filename);

	static bool isFunctionWord(string pos);
	static bool isClausalVerb(string verb);
	static void getFrameFeatures(string verb, vector<string>& strfeatures);
	static string getVerbLemma(string verb);

	static string getLemma(string word, string pos);
	static string getPolarity(string word, string lemma, string pos);
	static string getPolarStrength(string word, string lemma, string pos);
	static void updateOpLexicon(vector<string> &positive_words, vector<string> &pos_pos,
			vector<string> &negative_words, vector<string> &neg_pos);

	static bool HasVerbDerivedForm(string word, string pos);
	static bool IsWordnetSynonym(string word1, string pos1, string word2, string pos2);

	/*Set<String> reportVerb = new HashSet<String>(Arrays.asList(
    "accuse", "acknowledge", "add", "admit", "advise", "agree", "alert",
    "allege", "announce", "answer", "apologize", "argue",
    "ask", "assert", "assure", "beg", "blame", "boast",
    "caution", "charge", "cite", "claim", "clarify", "command", "comment",
    "compare", "complain", "concede", "conclude", "confirm", "confront", "congratulate",
    "contend", "contradict", "convey", "counter", "criticize",
    "debate", "decide", "declare", "defend", "demand", "demonstrate", "deny",
    "describe", "determine", "disagree", "disclose", "discount", "discover", "discuss",
    "dismiss", "dispute", "disregard", "doubt", "emphasize", "encourage", "endorse",
    "equate", "estimate", "expect", "explain", "express", "extoll", "fear", "feel",
    "find", "forbid", "forecast", "foretell", "forget", "gather", "guarantee", "guess",
    "hear", "hint", "hope", "illustrate", "imagine", "imply", "indicate", "inform",
    "insert", "insist", "instruct", "interpret", "interview", "invite", "issue",
    "justify", "learn", "maintain", "mean", "mention", "negotiate", "note",
    "observe", "offer", "oppose", "order", "persuade", "pledge", "point", "point out",
    "praise", "pray", "predict", "prefer", "present", "promise", "prompt", "propose",
    "protest", "prove", "provoke", "question", "quote", "raise", "rally", "read",
    "reaffirm", "realise", "realize", "rebut", "recall", "reckon", "recommend", "refer",
    "reflect", "refuse", "refute", "reiterate", "reject", "relate", "remark",
    "remember", "remind", "repeat", "reply", "report", "request", "respond",
    "restate", "reveal", "rule", "say", "see", "show", "signal", "sing",
    "slam", "speculate", "spoke", "spread", "state", "stipulate", "stress",
    "suggest", "support", "suppose", "surmise", "suspect", "swear", "teach",
    "tell", "testify", "think", "threaten", "told", "uncover", "underline",
    "underscore", "urge", "voice", "vow", "warn", "welcome",
    "wish", "wonder", "worry", "write"));

public final Set<String> nonWords = new HashSet<String>(Arrays.asList("mm", "hmm", "ahem", "um"));
public final Set<String> copulas = new HashSet<String>(Arrays.asList("is","are","were", "was","be", "been","become","became","becomes","seem","seemed","seems","remain","remains","remained"));
public final Set<String> quantifiers = new HashSet<String>(Arrays.asList("not","every","any","none","everything","anything","nothing","all","enough"));
public final Set<String> parts = new HashSet<String>(Arrays.asList("half","one","two","three","four","five","six","seven","eight","nine","ten","hundred","thousand","million","billion","tens","dozens","hundreds","thousands","millions","billions","group","groups","bunch","number","numbers","pinch","amount","amount","total","all","mile","miles","pounds"));
public final Set<String> temporals = new HashSet<String>(Arrays.asList(
    "second", "minute", "hour", "day", "week", "month", "year", "decade", "century", "millennium",
    "monday", "tuesday", "wednesday", "thursday", "friday", "saturday", "sunday", "now",
    "yesterday", "tomorrow", "age", "time", "era", "epoch", "morning", "evening", "day", "night", "noon", "afternoon",
    "semester", "trimester", "quarter", "term", "winter", "spring", "summer", "fall", "autumn", "season",
    "january", "february", "march", "april", "may", "june", "july", "august", "september", "october", "november", "december"));


public final Set<String> femalePronouns = new HashSet<String>(Arrays.asList(new String[]{ "her", "hers", "herself", "she" }));
public final Set<String> malePronouns = new HashSet<String>(Arrays.asList(new String[]{ "he", "him", "himself", "his" }));
public final Set<String> neutralPronouns = new HashSet<String>(Arrays.asList(new String[]{ "it", "its", "itself", "where", "here", "there", "which" }));
public final Set<String> possessivePronouns = new HashSet<String>(Arrays.asList(new String[]{ "my", "your", "his", "her", "its","our","their","whose" }));
public final Set<String> otherPronouns = new HashSet<String>(Arrays.asList(new String[]{ "who", "whom", "whose", "where", "when","which" }));
public final Set<String> thirdPersonPronouns = new HashSet<String>(Arrays.asList(new String[]{ "he", "him", "himself", "his", "she", "her", "herself", "hers", "her", "it", "itself", "its", "one", "oneself", "one's", "they", "them", "themself", "themselves", "theirs", "their", "they", "them", "'em", "themselves" }));
public final Set<String> secondPersonPronouns = new HashSet<String>(Arrays.asList(new String[]{ "you", "yourself", "yours", "your", "yourselves" }));
public final Set<String> firstPersonPronouns = new HashSet<String>(Arrays.asList(new String[]{ "i", "me", "myself", "mine", "my", "we", "us", "ourself", "ourselves", "ours", "our" }));
public final Set<String> moneyPercentNumberPronouns = new HashSet<String>(Arrays.asList(new String[]{ "it", "its" }));
public final Set<String> dateTimePronouns = new HashSet<String>(Arrays.asList(new String[]{ "when" }));
public final Set<String> organizationPronouns = new HashSet<String>(Arrays.asList(new String[]{ "it", "its", "they", "their", "them", "which"}));
public final Set<String> locationPronouns = new HashSet<String>(Arrays.asList(new String[]{ "it", "its", "where", "here", "there" }));
public final Set<String> inanimatePronouns = new HashSet<String>(Arrays.asList(new String[]{ "it", "itself", "its", "where", "when" }));
public final Set<String> animatePronouns = new HashSet<String>(Arrays.asList(new String[]{ "i", "me", "myself", "mine", "my", "we", "us", "ourself", "ourselves", "ours", "our", "you", "yourself", "yours", "your", "yourselves", "he", "him", "himself", "his", "she", "her", "herself", "hers", "her", "one", "oneself", "one's", "they", "them", "themself", "themselves", "theirs", "their", "they", "them", "'em", "themselves", "who", "whom", "whose" }));
public final Set<String> indefinitePronouns = new HashSet<String>(Arrays.asList(new String[]{"another", "anybody", "anyone", "anything", "each", "either", "enough", "everybody", "everyone", "everything", "less", "little", "much", "neither", "no one", "nobody", "nothing", "one", "other", "plenty", "somebody", "someone", "something", "both", "few", "fewer", "many", "others", "several", "all", "any", "more", "most", "none", "some", "such"}));
public final Set<String> relativePronouns = new HashSet<String>(Arrays.asList(new String[]{"that","who","which","whom","where","whose"}));
public final Set<String> GPEPronouns = new HashSet<String>(Arrays.asList(new String[]{ "it", "itself", "its", "they","where" }));
public final Set<String> pluralPronouns = new HashSet<String>(Arrays.asList(new String[]{ "we", "us", "ourself", "ourselves", "ours", "our", "yourself", "yourselves", "they", "them", "themself", "themselves", "theirs", "their" }));
public final Set<String> singularPronouns = new HashSet<String>(Arrays.asList(new String[]{ "i", "me", "myself", "mine", "my", "yourself", "he", "him", "himself", "his", "she", "her", "herself", "hers", "her", "it", "itself", "its", "one", "oneself", "one's" }));
public final Set<String> facilityVehicleWeaponPronouns = new HashSet<String>(Arrays.asList(new String[]{ "it", "itself", "its", "they", "where" }));
public final Set<String> miscPronouns = new HashSet<String>(Arrays.asList(new String[]{"it", "itself", "its", "they", "where" }));
public final Set<String> reflexivePronouns = new HashSet<String>(Arrays.asList(new String[]{"myself", "yourself", "yourselves", "himself", "herself", "itself", "ourselves", "themselves", "oneself"}));
public final Set<String> transparentNouns = new HashSet<String>(Arrays.asList(new String[]{"bunch", "group",
    "breed", "class", "ilk", "kind", "half", "segment", "top", "bottom", "glass", "bottle",
    "box", "cup", "gem", "idiot", "unit", "part", "stage", "name", "division", "label", "group", "figure",
    "series", "member", "members", "first", "version", "site", "side", "role", "largest", "title", "fourth",
    "third", "second", "number", "place", "trio", "two", "one", "longest", "highest", "shortest",
    "head", "resident", "collection", "result", "last"
}));
public final Set<String> stopWords = new HashSet<String>(Arrays.asList(new String[]{"a", "an", "the", "of", "at",
    "on", "upon", "in", "to", "from", "out", "as", "so", "such", "or", "and", "those", "this", "these", "that",
    "for", ",", "is", "was", "am", "are", "'s", "been", "were"}));

public final Set<String> notOrganizationPRP = new HashSet<String>(Arrays.asList(new String[]{"i", "me", "myself",
    "mine", "my", "yourself", "he", "him", "himself", "his", "she", "her", "herself", "hers", "here"}));
*/
};

#endif /* DICTIONARY_H_ */
