/*
 * SemiCRFLearner.cpp
 *
 *  Created on: Feb 15, 2014
 *      Author: bishan
 */

#include "SemiCRFLearner.h"
#include <fstream>

void SemiCRFLearner::BuildTagger() {
	if (encoder.feature_index->max_span_length <= 0) {
		int max_length = 1;
		for (int i = 0; i < trainset->sentences.size(); ++i) {
			for (int k = 0; k < trainset->sentences[i]->entities.size(); ++k) {
				if (trainset->sentences[i]->entities[k].len() > max_length) {
					if (trainset->sentences[i]->entities[k].label != "O")
						max_length = trainset->sentences[i]->entities[k].len();
				}
			}
		}
		encoder.feature_index->max_span_length = max_length;
	}

	//build features
	encoder.labelx.clear();
	for (int i = 0; i < trainset->sentences.size(); ++i) {
		CRFPP::SemiTagger* x = new CRFPP::SemiTagger();
		if (x->SentenceToSemiTagger(trainset->sentences[i], encoder.feature_index, true, useparse)) {
			int xid = encoder.labelx.size();
			x->set_thread_id(xid % encoder.thread_num);
			encoder.labelx.push_back(x);
			xid++;
			if (xid %1000 == 0) {
				cout<<"Loaded "<<encoder.labelx.size()<<" instances..."<<endl;
			}
		} else {
			delete x;
			x = NULL;
		}
	}

	cout<<"Loaded "<<encoder.labelx.size()<<" instances"<<endl;
}

void SemiCRFLearner::InitializeEncoder(CRFPP::FeatureIndex *p_feature_index) {
	encoder.feature_index = (CRFPP::EncoderFeatureIndex*)p_feature_index;
}

void SemiCRFLearner::TrainEncoder(string modelfile) {
	cout<<"Feature size before shrinking "<<encoder.feature_index->size()<<endl;
	encoder.feature_index->shrink(encoder.freq);
	cout<<"Feature size after shrinking "<<encoder.feature_index->size()<<endl;

	encoder.RebuildAlpha();

	if (sgd) {
		encoder.SGD_learn(modelfile, learningrate, decayRate, sgd_iteration);
	} else {
		encoder.learn(modelfile.c_str());
	}
}

void SemiCRFPredictor::LoadModel(string modelfile) {
	feature_index = new CRFPP::DecoderFeatureIndex();
	feature_index->open(modelfile.c_str());
}

void SemiCRFPredictor::NBestPrediction(string resultfile) {
	ifstream infile(resultfile.c_str(), ios::in);
	vector<string> results;
	string sent_result = "";
	string str;
	while (getline(infile, str)) {
		if (str == "") {
			results.push_back(sent_result);
			sent_result = "";
		} else {
			sent_result += str + "\n";
		}
	}
	infile.close();

	int n = 0;

	double alltruep = 0;
	double allpredictp = 0;
	double allpropp = 0;
	double allpropr = 0;

	for (int i = 0; i < results.size(); ++i) {
		vector<string> lines;
		Utils::Split(results[i], '\n', lines);
		vector<vector<string> > values;
		for (int k = 0; k < lines.size(); ++k) {
			vector<string> fields;
			Utils::Split(lines[k], '\t', fields);
			values.push_back(fields);
		}

		int ncol = values[0].size();
		if (ncol < 3) continue;

		// read gold standard labels
		vector<Span> answers;
		int index = 1;
		int p = 0;
		while (p < values.size()) {
			if (values[p][index] != "O") {
				int start = p;
				p++;
				while (p < values.size() && values[p][index] == values[start][index]) {
					p++;
				}
				int end = p-1;
				Span eval_en;
				eval_en.start = start;
				eval_en.end = end;
				eval_en.label = values[start][index];
				answers.push_back(eval_en);
			} else {
				p++;
			}
		}

		vector<vector<Span> > nbest_results(ncol-2);
		for (int k = 2; k < ncol; ++k) {
			int index = k;
			int p = 0;
			while (p < values.size()) {
				if (values[p][index] != "O") {
					int start = p;
					p++;
					while (p < values.size() && values[p][index] == values[start][index]) {
						p++;
					}
					int end = p-1;
					Span eval_en;
					eval_en.start = start;
					eval_en.end = end;
					eval_en.label = values[start][index];
					nbest_results[k-2].push_back(eval_en);
				} else {
					p++;
				}
			}
		}

		// pick the best overlapping prediction
		double truep = answers.size();
		double best_index = 0;
		double best_predictp = 0.0;
		double best_propr = 0.0;
		double best_propp = 0.0;
		double best_f1 = 0.0;
		for (int k = 0; k < nbest_results.size(); ++k) {
			double propr = 0.0;
			double propp = 0.0;
			if (answers.size() == 0 && nbest_results[k].size() == 0) continue;

			double predictp = nbest_results[k].size();
			ProportionMatch(answers, nbest_results[k], propr, propp);
			double pre = propp==0 ? 0 : propp/predictp;
			double rec = propr==0 ? 0 : propr/truep;
			double f = pre+rec==0 ? 0 : 2.0*pre*rec/(pre+rec);
			if (f > best_f1) {
				best_f1 = f;
				best_predictp = predictp;
				best_propr = propr;
				best_propp = propp;

				best_index = k;
			}
		}

		alltruep += truep;
		allpredictp += best_predictp;
		allpropp += best_propp;
		allpropr += best_propr;

		n++;
	}

	double pre = allpropp==0 ? 0 : allpropp/allpredictp;
	double rec = allpropr==0 ? 0 : allpropr/alltruep;
	double f = pre+rec==0 ? 0 : 2.0*pre*rec/(pre+rec);

	cout<<"pre: "<<pre<<" rec: "<<rec<<" f: "<<f<<endl;
}

void SemiCRFPredictor::TaggingPrediction(string resultfile) {
	// Evaluation
	cout<<"test instance size: "<<testx.size()<<endl;
	ofstream outfile(resultfile.c_str(), ios::out);

	// read normal results
	vector<vector<string> > answers;
	vector<vector<string> > results;

	for(int i=0; i<testx.size(); ++i) {
		Sentence *sent = testx[i]->sent;

		vector<string> labels(sent->size());
		fill(labels.begin(), labels.end(), "O");
		for (int k = 0; k < sent->entities.size(); ++k) {
			for (int j = sent->entities[k].start; j <= sent->entities[k].end; ++j) {
				labels[j] = sent->entities[k].label;
			}
		}

		testx[i]->Inference();

		// read results
		vector<string> pred_labels(sent->size());
		fill(pred_labels.begin(), pred_labels.end(), "O");
		int k = 0;
		while (k < testx[i]->result_.size()) {
			if (testx[i]->result_[k] != 0) {
				int start = k;
				k++;
				while (k < testx[i]->result_.size() && testx[i]->result_[k] == testx[i]->result_[start])
					k++;

				string label = feature_index->ystr(testx[i]->result_[start]);
				for (int j = start; j <= k-1; ++j) {
					pred_labels[j] = label;
				}
			} else {
				++k;
			}
		}

		answers.push_back(labels);
		results.push_back(pred_labels);

		// set prediction results in sentence
		outfile<<testx[i]->toString();
	}

	std::stringstream ss;
	ComputeScore(answers, results, eval_param, ss);

	outfile<<ss.str()<<endl;
	outfile.close();
}

void SemiCRFPredictor::BuildTagger() {
	testx.clear();

	for (int i = 0; i < testset->sentences.size(); ++i) {
		Sentence *sent = testset->sentences[i];
		CRFPP::SemiTagger* x = new CRFPP::SemiTagger();
		if (x->SentenceToSemiTagger(sent, feature_index, false)) {
			x->nbest_ = nbest;
			testx.push_back(x);
		} else {
			delete x;
			x = NULL;
		}
	}
}
