/*
 * SemiCRFLearner.h
 *
 *  Created on: Feb 15, 2014
 *      Author: bishan
 */

#ifndef SEMICRFLEARNER_H_
#define SEMICRFLEARNER_H_
#include "./CRF/semi_encoder.h"
#include "Dataset.h"
#include "Param.h"
#include "TaggingEval.h"

class SemiCRFLearner {
public:
	SemiCRFLearner() : trainset(NULL){
		useparse = false;

		sgd = false;
		adagrad = false;
		learningrate = 0.1;
		decayRate = 0.01;
	}
	~SemiCRFLearner() {}

	Dataset *trainset;

	bool useparse;

	bool sgd;
	int sgd_iteration;
	double learningrate;
	double decayRate;
	bool adagrad;

	CRFPP::SemiCRFEncoder encoder;

public:
    void initParam(Param param) {
		sgd = param.use_sgd_learn;
		sgd_iteration = param.sgd_iteration;
		learningrate = param.learningrate;
		useparse = param.use_parse;

		encoder.thread_num = param.threadnum;
		encoder.C = param.regularizer;
		encoder.freq = param.freq;
	}

    void InitializeEncoder(CRFPP::FeatureIndex *p_feature_index);

	void BuildTagger();

	void TrainEncoder(string modelfile);

};

class SemiCRFPredictor {
public:
	SemiCRFPredictor() {
		testset = new Dataset();
		feature_index = NULL;
		nbest = 1;
	}

	~SemiCRFPredictor() {
		for (int i = 0; i <testx.size(); ++i) {
			delete testx[i];
			testx[i] = NULL;
		}
	}
public:
	Dataset *testset;

	CRFPP::DecoderFeatureIndex *feature_index;
	vector<CRFPP::Tagger *> testx;
	int nbest;

	EvalParam eval_param;
public:
	void InitializeDecoder(CRFPP::FeatureIndex *p_feature_index) {
		feature_index = (CRFPP::DecoderFeatureIndex*)p_feature_index;
	}
	void LoadModel(string modelfile);

	void Prediction(string outputfile);
	void TaggingPrediction(string errfile);

	void BuildTagger();

	void NBestPrediction(string resultfile);
};


#endif /* SEMICRFLEARNER_H_ */
