#include "Sentence.h"

Sentence::~Sentence() {
	vector<string>().swap(words);
	vector<string>().swap(postags);
	vector<string>().swap(token_labels);
	vector<string>().swap(lemmas);

	if (penntree != 0) delete penntree;
	if (G != 0) delete G;
	if (dep_tree != 0) delete dep_tree;
}

string Sentence::ToString() {
	string str = "";
	for(int i=0; i<token_labels.size(); i++)
		str += token_labels[i]+" ";
	return str;
}

string Sentence::toCRFstring(set<string> &labelmap)
{
	string str = "";
	for(int i=0; i<words.size(); i++) {
		str += words[i]+"\t";
		str += postags[i]+"\t";
		//label
		if(token_labels[i] == "O") //"O"
			str += "O\n";
		else {
			string label = token_labels[i];
			if(label.substr(0,2) == "B_")
				label = label.substr(2);
			if(labelmap.find(label) != labelmap.end())
				str += token_labels[i]+"\n";
			else
				str += "O\n";
		}
	}

	return str;
}

void Sentence::parseSent(string sentstr)
{
	words.clear();
	postags.clear();
	token_labels.clear();

	vector<string> fields;
	Utils::Split(sentstr, ' ', fields);
	//remove "B_"
	for(int i=0; i<fields.size(); i++) {
		vector<string> values;
		Utils::Split(fields[i], '/', values);

		if (values.size() != 3) {
			cout<<"sentence file format error!"<<endl;
			cout<<sentstr<<endl;
			exit(0);
		}

		string word = values[0];
		string pos = values[1];
		assert(pos != "" && word != "");

		string lower_word = Utils::toLower(word);
		string lemma = Dictionary::getLemma(word, pos);
		words.push_back(word);
		postags.push_back(pos);
		lemmas.push_back(lemma);

		token_labels.push_back(values[2]);
	}

	sent = "";
	for(int i=0; i<words.size(); i++)
		sent += words[i]+" ";
}

bool Sentence::buildDepGraph(vector<string> &lines)
{
	vector<pair<int, int> > dep_pairs;
	vector<string> dep_relations;

	//map< int, vector<string> > depmap;
    G = NULL;
    G = new DependencyGraph();

    int max_wid = -1;
	for(int i=0; i<lines.size(); i++) {
	int sindex = lines[i].find("(");
	int eindex = lines[i].rfind(")");
	string dep = lines[i].substr(0, sindex);

	string pairs = lines[i].substr(sindex+1, eindex-sindex-1);
	vector<string> tmp;
	sindex = pairs.find(", ");
	tmp.push_back(pairs.substr(0, sindex));
	tmp.push_back(pairs.substr(sindex+2));
	assert(tmp.size() == 2);

	sindex = tmp[0].size()-1;
	while(tmp[0][sindex] != '-') sindex--;
	int wid1 = atoi(tmp[0].substr(sindex+1).c_str()); //wordid-1, since id 0 refers to ROOT

	sindex = tmp[1].size()-1;
	while(tmp[1][sindex] != '-') sindex--;
	int wid2 = atoi(tmp[1].substr(sindex+1).c_str());

	// Not including the root relation
	if (wid1 == 0) {
		continue;
	} else if (wid2 == 0) {
		continue;
	}
	wid1 = wid1 - 1;
	wid2 = wid2 - 1;
	if (wid1 > max_wid) {
		max_wid = wid1;
	}
	if (wid2 > max_wid) {
		max_wid = wid2;
	}

	dep_pairs.push_back(make_pair(wid1, wid2));
	dep_relations.push_back(dep);
	}
	if (max_wid+1 > words.size()) {
		return false;
	}

	if (!G->buildGraph(dep_pairs, dep_relations, words.size())) {
		delete G;
		return false;
	}

    //return G->buildGraph(depmap, words.size());
	return true;
}

bool Sentence::buildDepTree(vector<string> &lines)
{
	vector<pair<int, int> > dep_pairs;
	vector<string> dep_relations;

    dep_tree = NULL;
    dep_tree = new DependencyTree();

    int max_wid = -1;
    for(int i=0; i<lines.size(); i++) {
		int sindex = lines[i].find("(");
		int eindex = lines[i].rfind(")");
		string dep = lines[i].substr(0, sindex);

		string pairs = lines[i].substr(sindex+1, eindex-sindex-1);
		vector<string> tmp;
		sindex = pairs.find(", ");
		tmp.push_back(pairs.substr(0, sindex));
		tmp.push_back(pairs.substr(sindex+2));
		assert(tmp.size() == 2);

		sindex = tmp[0].size()-1;
		while(tmp[0][sindex] != '-') sindex--;
		int wid1 = atoi(tmp[0].substr(sindex+1).c_str()); //wordid-1, since id 0 refers to ROOT

		sindex = tmp[1].size()-1;
		while(tmp[1][sindex] != '-') sindex--;
		int wid2 = atoi(tmp[1].substr(sindex+1).c_str());

		// Not including the root relation
		if (wid1 == 0) {
			continue;
		} else if (wid2 == 0) {
			continue;
		}
		wid1 = wid1 - 1;
		wid2 = wid2 - 1;
		if (wid1 > max_wid) {
			max_wid = wid1;
		}
		if (wid2 > max_wid) {
			max_wid = wid2;
		}

		dep_pairs.push_back(make_pair(wid1, wid2));
		dep_relations.push_back(dep);
    }
    if (max_wid+1 > words.size()) {
    		return false;
    }

    if (!dep_tree->BuildTree(dep_pairs, dep_relations, words.size())) {
    		delete dep_tree;
    		return false;
    }

    G = new DependencyGraph();
    G->buildGraph(dep_pairs, dep_relations, words.size());
    return true;
}

bool Sentence::buildPennTree(vector<string> &lines)
{
	string text = "";
	for(int i=0; i<lines.size(); i++)
	{
		string str = lines[i];
		Utils::Trim(str);
		text += str;
	}
	penntree = new PennTree();
	penntree->ReadTree(text);

	return true;

}

void Sentence::buildPennTree(string parse_tree)
{
	if (penntree) {
		delete penntree;
		penntree = NULL;
	}
	penntree = new PennTree();
	penntree->ReadTree(parse_tree);
}

void Sentence::buildDiscourseTree(vector<string> &lines)
{
	string text = "";
	for(int i=0; i<lines.size(); i++)
	{
		string str = lines[i];
		Utils::Trim(str);
		text += str;
	}
	penntree = new PennTree();
	penntree->ReadDiscourseTree(text);
}

string Sentence::GetSpanText(int start, int end) {
	string str = "";
	for (int i = start; i <= end; ++i) {
		str += words[i] + " ";
	}
	return str.substr(0, str.size()-1);
}

string Sentence::GetContext(int start, int end) {
	int context_start = max(0, start - 3);
	int context_end = min(end + 3, (int)words.size() - 1);
	string str = "";
	for (int i = context_start; i < start; ++i) {
		str += words[i] + " ";
	}
	str += "[";
	for (int i = start; i <= end; ++i) {
		str += words[i] + " ";
	}
	str += "] ";

	for (int i = end + 1; i <= context_end; ++i) {
		str += words[i] + " ";
	}
	return str.substr(0, str.size()-1);
}
void Sentence::genUnigramFeatures(int start, int end, vector<string> &features) {
	//word sequence
	for(int i=start; i<end+1; i++) {
		features.push_back(words[i]);
	}
	//pos
	for(int i=start; i<end+1; i++)
		features.push_back(postags[i]);
}

bool Sentence::EmptySpan(int start, int end) {
	for (int i = start; i <= end; ++i) {
		if (token_labels[i] != "O") return false;
	}
	return true;
}

void Sentence::genSegmentBoundaryFeatures(int start, int end, vector<string>& features) {
	// boundary
	 genBoundaryFeatures(-1, start, features);
	 genBoundaryFeatures(1, end, features);
}

void Sentence::genSegmentSyntacticFeatures(int start, int end, vector<string>& features) {
	 // syntactic role
	string role = "";
	for (int i = start; i <= end; ++i) {
		role += postags[i] + "_";
	}
	features.push_back(role);

//	if (penntree != NULL) {
//		string role = penntree->GetSyntacticPath(start, end);
//		features.push_back(role);
//	}

	// syntactic parse
//	if (penntree != NULL) {
//		vector<ParseTreeNode *> parsenodes;
//		penntree->getSegmentInfo(start, end, parsenodes);
//		string str = "";
//		for (int i = 0; i < parsenodes.size(); ++i) {
//			str += parsenodes[i]->getValue()+"_";
//		}
//		features.push_back("Syn_"+str);
//	}
}

void Sentence::genSegmentFeatures(int start, int end, vector<string>& features)
{
	 // unigram
	 for (int cur = start; cur <= end; ++cur) {
		 vector<string> cur_features;
		 genBasicFeatures(cur, cur_features);

		 for (int i = 0; i < cur_features.size(); ++i) {
			 features.push_back(cur_features[i]);
		 }
	 }

	 // position
//	 if (start == 0) {
//		 features.push_back("isStart");
//	 }
//	 if (end == size()-1) {
//		 features.push_back("isEnd");
//	 }
}

void Sentence::genEmbeddingFeatures(int start, int end,
		map<string, int> &lookup_table,
		vector<vector<float> > &word_vec,
		vector<double>& feature_values)
{
	 if (lookup_table.size() == 0 || word_vec.size() == 0) return;

	 feature_values.resize(word_vec[0].size(), 0.0);

	 int c = 0;
	 for (int cur = start; cur <= end; ++cur) {
		 string word = words[cur];
		 if (lookup_table.find(word) != lookup_table.end()) {
			 int id = lookup_table[word];
			 for (int i = 0; i < word_vec[id].size(); ++i) {
				 feature_values[i] += word_vec[id][i];
			 }
			 c++;
		 } else {
			 if (lookup_table.find(lemmas[cur]) != lookup_table.end()) {
				 int id = lookup_table[lemmas[cur]];
				 for (int i = 0; i < word_vec[id].size(); ++i) {
					 feature_values[i] += word_vec[id][i];
				 }
				 c++;
			 }
		 }
	 }
	 // average word_vec
	 if (c > 0) {
		 for (int i = 0; i < feature_values.size(); ++i) {
			 feature_values[i] = (double)feature_values[i]/c;
		 }
	 }
}

void Sentence::genSpanFeatures(int start, int end, vector<string>& features)
{
	 // unigram
	 for (int cur = start; cur <= end; ++cur) {
		 genBasicFeatures(cur, features);
	 }

	 // bigram
	 for (int cur = start; cur < end; ++cur) {
		 string bigram = words[cur]+"_"+words[cur+1];
		 features.push_back(bigram);
		 if (postags.size() > 0) {
			 features.push_back(postags[cur]+"_"+postags[cur+1]);
		 }
	 }

	 // trigram
	 /*for (int cur = start; cur < end-1; ++cur) {
		 string trigram = words[cur]+"_"+words[cur+1]+"_"+words[cur+2];
		 features.push_back(trigram);
		 features.push_back(postags[cur]+"_"+postags[cur+1]+"_"+postags[cur+2]);
	 }*/
}

void Sentence::genBoundaryFeatures(int offset, int cur, vector<string> &features) {
	if (offset == 0 || cur < 0 || cur >= (int)words.size()) return;
	string prefix = Utils::int2string(offset);

	 int index = cur + offset;
	 if (index >= 0 && index < (int)words.size()) {
		 features.push_back(prefix+"_"+ words[index]);

		 string postag = "";
		 if (postags.size() > 0) {
			 postag = postags[index];
			 features.push_back(prefix+"_"+postag);
		 }

		 //if (lemmas.size() > 0)
		 //	 features.push_back(prefix+"_Lemma_"+ lemmas[index]);
	 }
}

void Sentence::genCRFWindowFeatures(int offset, int cur, vector<string> &features) {
	if (offset == 0 || cur < 0 || cur >= (int)words.size()) return;
	string prefix = Utils::int2string(offset);

	 int index = cur + offset;
	 if (index >= 0 && index < (int)words.size()) {
		 features.push_back(prefix+"_"+words[index]);

		 string postag = "";
		 if (postags.size() > 0) {
			 postag = postags[index];
			 features.push_back(prefix+"_"+postag);
		 }

//		 string lemma = "";
//		 if (lemmas.size() > 0) {
//			 lemma = lemmas[index];
//			 features.push_back(prefix+"_Lemma_"+lemma);
//		 }
	 }
}

void Sentence::genBasicFeatures(int cur, vector<string> &features) {
	string word = words[cur];

	// unigram
	features.push_back(word);

	// postag
	string postag = "";
	if (postags.size() > 0) {
		postag = postags[cur];
		features.push_back(postag);
	}

//	string lemma = "";
//	if (lemmas.size() > 0) {
//		features.push_back(lemmas[cur]);
//	}
}

void Sentence::genDepFeatures(int cur, vector<string> &features) {
	GraphNode* node = G->nodes[cur];
	for(int i=0; i<node->getEdgeNum(); i++) {
		int arg = node->getEdge(i)->rv->pos;
		// dep + subj
		string dep = node->getEdge(i)->value;
		if (dep == "neg_-1" || dep == "ccomp_+1") {
			string subj = Dictionary::getPolarStrength(words[arg], lemmas[arg], postags[arg]);
			if (subj != "") {
				string fstr = dep+"_"+subj;
				if (fstr == "neg_-1_strongsubj" || fstr == "ccomp_+1_weaksubj")
					features.push_back("Dep_"+fstr);
			}
		}

		// dep + pos
		string fstr = dep+"_" + postags[arg];
		if (fstr == "advmod_+1_DT" || fstr == "rcmod_-1_FW" || fstr == "ccomp_+1_NN" ||
				fstr == "ccomp_+1_MD") {
			features.push_back("Dep_"+fstr);
		}
	}
}

void Sentence::genParseFeatures(int cur, vector<string> &features) {
	ParseTreeNode* node = penntree->leaves[cur];
	assert(node->getStartPos() == cur);
	ParseTreeNode *parent = (ParseTreeNode *)node->GetParent();
	if (parent != NULL) {
		if (parent->getValue() == "VP") {
			features.push_back("Parent_"+parent->getValue());
		}
	}
}


void Sentence::GenPathFeatures(int pi, int ci, vector<string> &features) {
	//if (dep_tree->nodes[pi]->observed && dep_tree->nodes[pi]->observed) {
	//	return;
	//}

	//features.push_back("B_");
	//features.push_back("Dep_"+dep_tree->nodes[ci]->parent_dep);

	string pobserve = "";
	string cobserve = "";
	string dep = dep_tree->nodes[ci]->parent_dep;

	/*if (dep_tree->nodes[pi]->observed) {
		pobserve = "true";
		features.push_back("Parent_observe");
	}
	if (dep_tree->nodes[ci]->observed) {
		cobserve = "true";
		features.push_back("Child_observe");
	}*/

	features.push_back("Dep_"+dep);

	features.push_back("Parent_"+words[pi]);
	features.push_back("Parent_"+postags[pi]);
	if (lemmas.size() > pi)
		features.push_back("Parent_"+lemmas[pi]);
	features.push_back("Child_"+words[ci]);
	features.push_back("Child_"+postags[ci]);
	if (lemmas.size() > ci)
		features.push_back("Child_"+lemmas[ci]);

	//features.push_back("B_");
}
